/* FILE: process.c              Copyright (c), Symmetric Research 2001-2006
 *
 * Watch for a PARxCH .out file, calling a threshold detector when it
 * appears.
 *
 * If the threshold exceeds the specified value, the .out is renamed
 * with a timestamp filename, otherwise deleted.
 *
 * Note that the processing is put to sleep for intervals to give other
 * processes a chance to execute.
 *
 *
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>        // required for O_RDONLY define
#include <time.h>         // required for time function proto

#include "parxch.h"       // required for simp.h ...
#include "simp.h"         // required for the HDR layout ...
#include "srhelper.h"     // required for the SR helper functions



/* REV MESSAGE:
 *
 * 06/20/05 Added ChopBuffer so Linux won't complain about fwritable-strings
 * 01/01/04 Previous rev
 *
 */

void Rev( void ) {

        // Print the rev info. 

        printf( "%s process : (c) SR 2006 : Rev {03/20/06}\n", PARXCH_MODELNAME );

}





/* ERRORS:
 *
 *
 *
 *
 */

char *UsageMsg[] = {
        "                                                          ",
        "Usage: process hdrfile outfile threshold                  ",
        "                                                          ",
        "where                                                     ",
        "      hdrfile   = binary .hdr file from simp/scope        ",
        "      outfile   = binary .out file from simp/scope        ",
        "      threshold = abs integer threshold level (eg 12345)  ",
        NULL,
        };

void Error( char *msg ) {

        int i;
        char *Chop( char *s );  // see below for code ... 


        // Display the rev. 

        Rev();


        // Print the error and usage message and quit. 

        printf( "ERROR: %s\n", msg );
        
        for ( i = 0 ; UsageMsg[i] ; i++ )
                printf( "%s", Chop( UsageMsg[i] ) );
        
        exit( 1 );

}

static char ChopBuffer[256];

char *Chop( char *s ) {

        char *p;


        // Chop spaces ' ' off the end of a string and add '\n'. 

        strncpy( ChopBuffer, s, 255 );

        p = ChopBuffer;

        while ( *p )

                p++;


        if ( *--p == '\n' ) {

                while ( *--p == ' ' )

                        ; // back up

                *++p = '\n';
                *++p = '\0';

                }

        return( ChopBuffer );
}




/* HDR FILE READ:
 *
 *
 *
 */

struct HDR_data HDR;

void ReadHdrFile( char *HdrFilename ) {

        FILE *FpHdrFile;



        // Read in the .hdr data file.

        FpHdrFile = fopen( HdrFilename, "rb" );

        if ( !FpHdrFile )

                Error( "unable to open .hdr file" );

        fread( &HDR, 1, sizeof( HDR ), FpHdrFile );

        fclose( FpHdrFile );



        // Examine the Id string. */

        if ( strncmp( HDR.Id, "SR PAR", 6 ) != 0 ) 

                Error( "invalid .hdr file id" );

}






/* OUT DATA FILE READ:
 *
 * The .out file is processed in chunks.  This function reads in a buffer
 * of nvalues.
 *
 *
 *
 */

int GetOutData( FILE *FpOutFile, long *values, int nvalues, int *nread ) {

        int nbytes;


        nbytes = fread( values, (int)HDR.WordSize, nvalues, FpOutFile );

        if ( nbytes )

                { *nread = (int)(nbytes/HDR.WordSize); return( 1 ); }

        else
                return( 0 );

}







/* DATA BUFFER:
 *
 * Buffer for processing the .out data file.
 *
 *
 *
 */

#define NVALUES         5000

long VALUES[ NVALUES ];





/* PARxCH ABSOLUTE VALUE:
 *
 * The PARxCH drivers return 24 bit signed or offset values.  It is assumed
 * the program is running the signed default.
 *
 * This function computes the absolute value of the 24 bit twos compliment
 * signed PARxCH values.
 *
 *
 *
 */

unsigned long ParXchAbs( long v ) {


        // Compute the absolute value of a 24 bit PARxCH signed value.
        // Note: negative values have already been signed extended to 32 bits.

        if ( (unsigned long)v > 0x007FFFFF )

                v = -v;

        return( v );

}






/* THRESHOLD DETECTOR:
 *
 * This function compares the entire .out file against the threshold.
 *
 * Even when the function finds the threshold is exceeded, it continues
 * on to scan all the data in the file so it runs in a predictable amount
 * of time.
 *
 * Note that the data is actually multiplexed, but it doesn't matter for
 * a simple threshold scan.
 *
 * It is assumed the PARxCH DataCoding is SIGNED.
 *
 *
 *
 *
 */

#define EXCEEDED                1
#define NOT_EXCEEDED            0

int CheckThreshold( FILE *FpOutFile, unsigned long Threshold ) {

        int status, nread, i;


        // Compare the .out file against the threshold.

        status = NOT_EXCEEDED;

        while ( GetOutData( FpOutFile, VALUES, NVALUES, &nread ) ) {

                for ( i = 0 ; i < nread ; i++ ) {

                        if ( ParXchAbs( VALUES[i] ) >= Threshold )

                                status = EXCEEDED;
                        }
                }


        return( status );

}








/* MAIN:
 *
 *
 *
 *
 */

int main( int argc, char *argv[] ) {

        int           status;
        unsigned long OutFileSize, Threshold;
        FILE         *FpOutFile;
        char         *OutFilename, *HdrFilename, *StrThreshold, Newname[32];


        // Allow getchar to return even if a key has not been pressed.

        if ( SrKeyboardNonBlock( ) == -1 )
               Error( "Can't set non-blocking read ..." );



        // Get the command line arguments.

        if ( argc != 4 )

                Error( "wrong number of command line arguments" );

        HdrFilename  = argv[1];
        OutFilename  = argv[2];
        StrThreshold = argv[3];





        // Convert the command line threshold argument to an integer value.

        if ( sscanf( StrThreshold, "%ld", &Threshold ) != 1 )

                Error( "invalid threshold value" );



        printf( "Starting threshold test.  Press ENTER to exit.\n\n" );


        // Wait for the .out file to appear.

        printf( "Waiting for %s data file to appear ...\n", OutFilename );

polling_loop:



        // Try to open it.

        FpOutFile = fopen( OutFilename, "rb" );


        if ( !FpOutFile /* == file not found */ ) {


                // Give the user a chance to quit ...

                if ( SrGetKeyCheck() != EOF ) {
                        printf( "Processing stopped.\n" );
                        exit( 0 );
                        }


                else // sleep for awhile and continue polling.

                        { SrSleep( 100 /* = time in ms */ ); goto polling_loop;}

                }
                






        // Found the .out file, start processing ...





        // Read the header file.

        ReadHdrFile( HdrFilename );


        // Check .out filesize against .hdr info.

        OutFileSize = HDR.Channels * HDR.RecordPts * HDR.WordSize;

        if ( (unsigned long)SrFileLength( FpOutFile ) != OutFileSize )

                Error( ".out filesize != .hdr info" );






        // Call the threshold detector, renaming if exceeded.

        printf( "Found %s, thresh test ... ", OutFilename );


        status = CheckThreshold( FpOutFile, Threshold );

            fclose( FpOutFile );



        if ( status == EXCEEDED ) {

                sprintf( Newname, "%08lX.OUT", time( NULL ) );

                    rename( OutFilename, Newname );

                printf( "EXCEEDED, renamed to %s\n", Newname );

                }

        else
                printf( "not exceeded\n" );






        // Delete the source .out file.

        remove( OutFilename );



        // Continue polling.

        goto polling_loop;


        return( 0 );
}
