/* FILE: dat2asc.c                Copyright (c), Symmetric Research, 2002-2008
 *      
 * File for converting a binary .dat file to ASCII.  GPS support is
 * optional, but is needed for converting the GPS PPS and serial data
 * records.  Be sure the PARGPS software is installed and define the
 * GPS_AVAILABLE macro when compiling to include GPS support.
 *
 * Change history:
 *
 *  2008-10-15  WCT  Support new SRDAT_SOURCE define and include invalid warning
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>       // for toupper
#include <time.h>


#include "srdat.h"
#include "srhelper.h"    // for SrFileExist


// If PARGPS software is available, the PPSTDATA and SERIALDATA 
// structures can be converted.  Otherwise they are skipped.

#ifdef GPS_AVAILABLE
#include "pargps.h"     // for PPSTDATA, SERIALDATA, TIMESTAMP, SR64BIT, etc
extern char ParGpsLastErrorMessage[];

#else // no GPS_AVAILABLE
                            // Method for computing time at an arbitrary sample
#define TIME_METHOD_NONE 0  //   Assume perfect sampling
#define TIME_METHOD_SPS  1  //   From timestamp and sampling rate
#define TIME_METHOD_GPS  2  //   From PARGPS and PC counter
#define TIME_METHOD_OBC  3  //   From PARGPS and PAR8CH 800 ns On Board Counter
#define TIME_METHOD_PCT  4  //   From PC system time

#define TS_VALID_MOST    0xB // See pargps.h for these.  Most is analog + pps + 
#define TS_VALID_ALL     0xF // obc, All includes serial too.

typedef int PPSDATA;
typedef int PPSTDATA;
typedef int SERIALDATA;
typedef struct _TIME_STAMP {  // this is a cut down version
        long   Sample;
        double SecSince1970;
        } TIMESTAMP;

int ParGpsCalcTimeZoneCorr( long StartTime );
int ParGpsSecTimeSplit( double Time, int *Year, int *Month, int *Day,
                        int *Hour, int *Minute, int *Second, long *Microsecond );
int ParGpsSecTimeCombine( int Year, int Month, int Day,
                          int Hour, int Minute, int Second, 
                          long MicroSecond, double *Time );
#endif // GPS_AVAILABLE


typedef struct SrDatTimeData TIMEPT;


// Prototypes

int   ProcessCommandLine( int argc, char** argv );
void  ShowHdr( FILE *fout, SRDAT_HDR *Hdr );
void  ShowTag( FILE *fout, SRDAT_TAG *Tag );
void  ShowData( FILE *fout, SRDAT_HDR *Hdr, void *Buffer,
                long DataType, unsigned long Nbytes,
                int HexFormat, int TimeFormat, int TimeMethod, 
                int ShowPcTime, int ShowAllInfo );
char *PickTimeTypeString( long Type );
char *PickSourceString( long Source );
char *PickDataCodingString( long DataCoding );
char *PickAtodModelString( long AtodModel );
char *PickGpsModelString( long GpsModel );
char *PickCounterTypeString( long Type );
char *PickFileNameTypeString( long FileNameType );
char *PickTriggerTypeString( long TriggerType );
char *PickTriggerParm1String( long TriggerType, long TriggerParm1 );
char *PickTriggerParm2String( long TriggerType, long TriggerParm2 );
char *PickTimeMethodString( long TimeMethod );
int   ReadDatRecord( SRDAT_HDR *Hdr,
                     void  *PC_BUFFER, unsigned int *npts,
                     void *PPS_BUFFER, unsigned int *npps,
                     void *SER_BUFFER, unsigned int *nser
                 );
int   SelectTimeMethod( SRDAT_HDR *Hdr );
void  Usage( char *msg );
void  ErrorWarning( char *msg, int status );
void  ErrorNotice( char *msg );
void  CleanExit( int status );

int InitTimeStamps( void );
int ComputeTime( SRDAT_HDR *Hdr, int TimeMethod, int TimeFormat,
                 long Sample, char *TimeStr );
int UpdateTimeStamps( long Sample );
int FillTimeStr( int TimeFormat, double Time, char *TimeStr );





// Global defines and data

char* UsageMsg[] = {
    "USAGE: dat2asc filename.DAT [/seq | /seqhex] [/hex] [/bare] [/tymd | /tsec] [/?]",
    "                                                                       ",
    "Dat2asc converts the specified .DAT file to ASCII text format.  The    ",
    "output filename is the same as the input filename except the suffix is ",
    "changed from .DAT to .ASC.  If the input file prefix contains a number,",
    "it can be treated as the first of a sequence where dat2asc repeately   ",
    "increments the file prefix and converts the next file.  The optional   ",
    "/seq and /seqhex arguments means dat2asc should increment the prefix   ",
    "in decimal (base 10) or in hex (base 16) respectively.  The optional   ",
    "/hex argument causes the analog data to be written in hex instead of   ",
    "decimal format. The /bare option shows only the analog data values and ",
    "none of the header or gps info.  The /tymd and /tsec options add an    ",
    "extra column for each sample showing time as Y/M/D H:M:S or seconds    ",
    "since 1970 respectively.  Use /? to show this usage message.           ",
    "                                                                       ",
    "EXAMPLES: dat2asc SIMP4CH.DAT                                          ",
    "          dat2asc 00000010.DAT /seq /hex /tymd                         ",
    NULL
    };



#define SEQTYPE_NONE  0
#define SEQTYPE_DEC   1
#define SEQTYPE_HEX   2

#define TIME_FORMAT_NONE  0
#define TIME_FORMAT_YMD   1
#define TIME_FORMAT_SEC   2

#define MARKCHAR_NONE  ' '
#define MARKCHAR_FOUND '<'

#define MAXDATAFMT     10
#define MAXTIMESTR     30
#define MAXFILENAME   256
#define MAXSTR        256
#define MAXCONVERT  10000

// MAXANALOG is defined large enough to hold data from any PARxCH from
// the PAR1CH to the PAR8CH.  The PAR1CH can have 2 channels, and the
// PAR4CH can have 5, but the PAR8CH can have up to 11 eg 8 analog + 1
// digital + 1 gps + 1 counter.

#define MAXANALOG (128*11) // 128 pts/ch * 11 channels 
#define MAXPPS        26   // This is large enough for sampling as often
#define MAXSER        26   // as about 5 samples per second.


int        TimeMethod, ShowAllInfo, ShowPcTime, TsValid;
int        SeqType, HexFormat, TimeFormat, Sequential;
long       CurrentPt16, CurrentPt24, SeqNum;
double     ObcCountsPerSample, BestCountsBetweenPps, CountsPerSecond;
char       FileIn[MAXFILENAME], FileOut[MAXFILENAME];
char       Buffer[MAXCONVERT], Msg[MAXSTR];
long       PC_BUFFER[MAXANALOG];
PPSTDATA   PPS_BUFFER[MAXPPS];
SERIALDATA SER_BUFFER[MAXSER];
TIMESTAMP  TimeStamp[2];
SRDAT_HDR  Hdr;
SRDAT_TAG  Tag;
FILE      *Fin, *Fout, *Fts;


//------------------------------------------------------------------------------
// ROUTINE: MAIN (dat2asc)
// PURPOSE: To convert binary .dat data files into ASCII.  See UsageMsg above
//          for a description of the command line arguments.
//------------------------------------------------------------------------------
int main( int argc, char** argv ) {

        unsigned int nbytes;


        // Announce program version.

        printf( "DAT2ASC [/?]: (c) SR 2008 : Rev {10/15/08}\n" );


        // Process command line arguments

        ProcessCommandLine( argc, argv );
        printf( "\n" );



        // Init file handles

        Fin = Fout = Fts = NULL;


        // If a time option was requested, init the timestamps

        if (TimeFormat != TIME_FORMAT_NONE) {
                if ( !InitTimeStamps() ) {
                        TimeFormat = TIME_FORMAT_NONE;
                        ErrorNotice( "Could not initialize time stamps.  Removing time column." );
                        }
                }
        

        // Loop until no more dat files requested or found

        printf( "Converting to ASCII:\n" );

        while ( 1 ) {

                // First check for existance of input file

                if ( !SrFileExist( FileIn ) ) {
                        sprintf( Msg, "No more files to process.  Stopping at %s.", FileIn );
                        ErrorWarning( Msg, 0 );
                        }
                
                // Open input file

                Fin = fopen( FileIn, "rb" );
                if (!Fin) {
                        sprintf( Msg, "Can not open input data file %s", FileIn );
                        ErrorWarning( Msg, 1 );
                        }




                // Read header

                printf( "Reading file %s ...\n", FileIn );

                if ( !SrDatHdrRead( Fin, &Hdr ) )
                        ErrorWarning( "Could not read .DAT header", 1 );
                
                if ( !SrDatHdrValidate( &Hdr ) )
                        ErrorWarning( "This is not a valid .DAT file", 1 );

                if ( Hdr.AtodSamplingRate <= 0.000001 ) {
                        printf( "WARNING: Bad sampling rate.  Reset to 1 sps.\n" );
                        Hdr.AtodSamplingRate = (double)1.0;
                        }



                // Open output file

                Fout = fopen( FileOut, "w+" );
                if (Fout == NULL) {
                        sprintf( Msg, "Can not open output ASCII file %s", FileOut );
                        ErrorWarning( Msg, 1 );
                        }


                // Write ASCII header

                printf( "Writing file %s ...\n", FileOut );

                if (ShowAllInfo) {
                        fprintf( Fout, "Converted from file %s\n\n", FileIn );
                        ShowHdr( Fout, &Hdr );
                        }


                // Initialize point counts that will be updated by ShowData
                
                CurrentPt16 = CurrentPt24 = Hdr.FileStartPtNum;

                
                
                // Read and write data records until reaching the end

                if ( !SrDatTagRead( Fin, &Tag ) ) {
                        ErrorNotice( "Could not read .DAT tag (main1)" );
                        Tag.TagId  = SRDAT_TAGID_EOF;
                        Tag.Nbytes = 0;
                        }

                if (ShowAllInfo)
                        ShowTag( Fout, &Tag );
                
                while (Tag.TagId != SRDAT_TAGID_EOF) {

                        nbytes = (unsigned int)Tag.Nbytes;
                        if ( (unsigned long)nbytes != Tag.Nbytes )
                                printf( "Warning data exceeds size allowed by unsigned int.\n");

                        if (Tag.Nbytes < MAXCONVERT) {
                                if ( !SrDatDataRead( Fin, Buffer, nbytes ) )
                                        ErrorWarning( "Could not read .DAT data", 1 );
                                
                                ShowData( Fout,
                                          &Hdr,
                                          Buffer,
                                          Tag.TagId,
                                          Tag.Nbytes,
                                          HexFormat,
                                          TimeFormat,
                                          TimeMethod,
                                          ShowPcTime,
                                          ShowAllInfo);
                                }
                        else {
                                SrDatDataSkip( Fin, nbytes );
                                printf("Exceeded allocated buffer size with %lu\n",
                                       Tag.Nbytes);
                                }


                        if ( !SrDatTagRead( Fin, &Tag ) ) {
                                ErrorNotice( "Could not read .DAT tag (main2)" );
                                Tag.TagId  = SRDAT_TAGID_EOF;
                                Tag.Nbytes = 0;
                                }

                        if (ShowAllInfo)
                                ShowTag( Fout, &Tag );
                        }



                // Close current files

                fclose( Fin );    Fin  = NULL;
                fclose( Fout );   Fout = NULL;


                
                // If only a single file was requested, we're done
                
                if (!Sequential)
                        break;

                // Otherwise, compute the next set of sequential file names
                
                else {
                        SeqNum++;
                        if ( SeqType == SEQTYPE_HEX ) {
                                sprintf( FileIn,  "%08lX.DAT", SeqNum );
                                sprintf( FileOut, "%08lX.ASC", SeqNum );
                                }
                        else { // ( SeqType == SEQTYPE_DEC )
                                sprintf( FileIn,  "%08ld.DAT", SeqNum );
                                sprintf( FileOut, "%08ld.ASC", SeqNum );
                                }
                        }

                }  // end while 1 (more files to process)

        return( 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Extract information provided by the user on the program command line.
//          See the UsageMsg structure defined above for allowed options.
//------------------------------------------------------------------------------
int ProcessCommandLine( int argc, char** argv ) {

        int   i, iprefix, len, useSeq, datFileArg;
        long  ltmp;
        char *p, FilePrefix[MAXFILENAME], PrefixCheck[MAXFILENAME];


        // Set defaults (global and local)

        SeqType     = SEQTYPE_NONE;
        HexFormat   = 0;
        ShowPcTime  = 0;
        ShowAllInfo = 1;
        TimeFormat  = TIME_FORMAT_NONE;

        useSeq      = 0;
        datFileArg  = 0;

        
        // Process command line arguments
        
        for ( i = 1 ; i < argc ; i++ ) {

                if (strcmp("/?",argv[i]) == 0)
                        Usage( "?" );

                else if (strcmp("/seq",argv[i]) == 0) {
                        if (useSeq == 1)
                                Usage( "Only one of /seq and /seqhex is allowed" );
                        useSeq = 1;
                        SeqType = SEQTYPE_DEC;
                        }

                else if (strcmp("/seqhex",argv[i]) == 0) {
                        if (useSeq == 1)
                                Usage( "Only one of /seq and /seqhex is allowed" );
                        useSeq = 1;
                        SeqType = SEQTYPE_HEX;
                        }

                else if (strcmp("/hex",argv[i]) == 0)
                        HexFormat = 1;

                else if (strcmp("/bare",argv[i]) == 0)
                        ShowAllInfo = 0;

                else if (strcmp("/tymd",argv[i]) == 0) {
                        if (TimeFormat != TIME_FORMAT_NONE)
                                Usage( "Only one of /tymd and /tsec is allowed" );
                        TimeFormat = TIME_FORMAT_YMD;
                        }

                else if (strcmp("/tsec",argv[i]) == 0) {
                        if (TimeFormat != TIME_FORMAT_NONE)
                                Usage( "Only one of /tymd and /tsec is allowed" );
                        TimeFormat = TIME_FORMAT_SEC;
                        }

                else if (strcmp("/ntp",argv[i]) == 0)  // This is a secret option
                        ShowPcTime = 1;                // for debugging use
                
                else if (datFileArg == 0)
                        datFileArg = i;

                else
                        Usage( "Bad command line argument" );
                }

        if (datFileArg == 0) // Did not find any .dat file name
                Usage( "Command line must include a .dat data file name" );

        if ( (argc < 2) || (argc > 6) )
                Usage( "Wrong number of command line arguments.  Must be 1 to 5." );


        
        // Fill some file name parameters
        
        strncpy( FileIn, argv[datFileArg], MAXFILENAME-1 );
        FileIn[MAXFILENAME-1] = '\0';


        // Find the end of the input file prefix

        len = strlen( FileIn );
        iprefix = -1;
        for ( i = len ; i > 0 ; i-- ) {
                if (FileIn[i-1] == '.') {
                        iprefix = i-1;
                        break;
                        }
                }
        
        if (iprefix < 0 || iprefix > len || iprefix > MAXFILENAME-4) {
                sprintf( Msg, "File name %s is invalid", FileIn );
                Usage( Msg );
                }


        // Extract the prefix, null terminate, and uppercase

        strncpy( FilePrefix, FileIn, iprefix );
        FilePrefix[iprefix] = '\0';

        p   = FilePrefix;
        len = strlen(FilePrefix);
        for ( i = 0 ; i < len ; i++ ) {
                *p = (char)toupper( *p );
                *p++;
                }



        // If optional sequential argument was found, determine starting 
        // sequential number from prefix and verify.

        Sequential = 0;
        SeqNum     = -1;

        if ( useSeq ) {
                        
                Sequential = 1;

                if (SeqType == SEQTYPE_HEX) {
                        if ( sscanf(FilePrefix, "%lX", &ltmp) == 1 )
                                SeqNum = ltmp;

                        sprintf( PrefixCheck, "%08lX", SeqNum );
                        if ( strcmp( FilePrefix, PrefixCheck ) != 0 )
                                Usage( "Can't set sequential hex value from filename" );
                        }
                
                else { // (SeqType == SEQTYPE_DEC)
                        if ( sscanf(FilePrefix, "%ld", &ltmp) == 1 )
                                SeqNum = ltmp;

                        sprintf( PrefixCheck, "%08ld", SeqNum );
                        if ( strcmp( FilePrefix, PrefixCheck ) != 0 )
                                Usage( "Can't set sequential decimal value from filename" );
                        }

                } // end useSeq



        // Make ASCII output name

        sprintf( FileOut, "%s.ASC", FilePrefix );

        
        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ShowHdr
// PURPOSE: Write header information to the ASCII output file.
//------------------------------------------------------------------------------
void ShowHdr( FILE *fout, SRDAT_HDR *Hdr ) {

        int i;
        char cDatId[5];

        
        // Check header size
        
        if (sizeof(*Hdr) != SRDAT_HDR_LAYOUTSIZE)
                ErrorWarning( "Header size not right", 1 );

        
        // Show header info
        
        strncpy( cDatId,((char *)(&Hdr->DatId)), 4);
        cDatId[4] = '\0';
        

        fprintf(fout, "Header:\n");

        fprintf(fout, "DatId               = 0x%lX %s\n", Hdr->DatId, cDatId);
        fprintf(fout, "DatRev              = %ld\n", Hdr->DatRev);
        fprintf(fout, "DatProduct          = %s\n",  Hdr->DatProduct);
        fprintf(fout, "DatApplication      = %s\n",  Hdr->DatApplication);
        fprintf(fout, "DatAtodModel        = %ld", Hdr->DatAtodModel);
        fprintf(fout, PickAtodModelString(Hdr->DatAtodModel) );
        fprintf(fout, "DatGpsModel         = %ld", Hdr->DatGpsModel);
        fprintf(fout, PickGpsModelString(Hdr->DatGpsModel) );
	
        fprintf(fout, "FileName            = %s\n",  Hdr->FileName);
        fprintf(fout, "FileCreateTimeHi    = %ld\n", Hdr->FileCreateTimeHi);
        fprintf(fout, "FileCreateTime      = %ld\n", Hdr->FileCreateTime);
        fprintf(fout, "FileSeqNum          = %ld\n", Hdr->FileSeqNum);
        fprintf(fout, "FileStartPtNum      = %ld\n", Hdr->FileStartPtNum);
        fprintf(fout, "FileNameType        = %ld",   Hdr->FileNameType);
        fprintf(fout, PickFileNameTypeString(Hdr->FileNameType) );
        fprintf(fout, "FileNameInc         = %ld\n", Hdr->FileNameInc);

        fprintf(fout, "ChannelsTotal       = %ld\n", Hdr->Channels);
        fprintf(fout, "ChannelsAnalog      = %ld\n", Hdr->ChannelsAnalog);
        fprintf(fout, "ChannelsDigital     = %ld\n", Hdr->ChannelsDigital);
        fprintf(fout, "ChannelsGps         = %ld\n", Hdr->ChannelsGps);
        fprintf(fout, "ChannelsCounter     = %ld\n", Hdr->ChannelsCounter);
        fprintf(fout, "ChannelDigIndex     = %ld\n", Hdr->ChannelDigIndex);
        fprintf(fout, "ChannelMarkIndex    = %ld\n", Hdr->ChannelMarkIndex);
        fprintf(fout, "ChannelObcIndex     = %ld\n", Hdr->ChannelObcIndex);
        fprintf(fout, "ChannelPtsPerRecord = %ld\n", Hdr->ChannelPtsPerRecord);
        fprintf(fout, "ChannelPtsPerFile   = %ld\n", Hdr->ChannelPtsPerFile);

	for ( i = 0 ; i < Hdr->Channels ; i++ ) {
		fprintf(fout, "Channel %02ld          = %s\n",
                          Hdr->ChannelList[i].Channel,
                          Hdr->ChannelList[i].Title );
                }

        fprintf(fout, "DigitalPerChan      = %ld\n", Hdr->DigitalPerChan);
	for ( i = 0 ; i < Hdr->DigitalPerChan ; i++ ) {
		fprintf(fout, "Digital %02ld          = %s\n",
                          Hdr->DigitalList[i].Channel,
                          Hdr->DigitalList[i].Title );
                }
        
        fprintf(fout, "AtodWordSize        = %ld\n", Hdr->AtodWordSize);
        fprintf(fout, "AtodDataCoding      = %ld",   Hdr->AtodDataCoding);
        fprintf(fout, PickDataCodingString(Hdr->AtodDataCoding) );
        fprintf(fout, "AtodSamplingRate    = %lf\n", Hdr->AtodSamplingRate);
        fprintf(fout, "AtodSpsMeasured     = %lf\n", Hdr->AtodSamplingRateMeasured);
        fprintf(fout, "AtodGain            = %lf\n", Hdr->AtodGain);
        fprintf(fout, "AtodFilterCoeff     = %lf\n", Hdr->AtodFilterCoeff);
        fprintf(fout, "AtodFilterScale     = %lf\n", Hdr->AtodFilterScale);

        fprintf(fout, "TriggerType         = %ld",   Hdr->TriggerType);
        fprintf(fout, PickTriggerTypeString(Hdr->TriggerType) );
        fprintf(fout, "TriggerParm1        = %ld", Hdr->TriggerParm1);
        fprintf(fout, PickTriggerParm1String(Hdr->TriggerType, Hdr->TriggerParm1) );
        fprintf(fout, "TriggerParm2        = %ld", Hdr->TriggerParm2);
        fprintf(fout, PickTriggerParm2String(Hdr->TriggerType, Hdr->TriggerParm2) );
        
        fprintf(fout, "CounterType         = %ld", Hdr->CounterType);
        fprintf(fout, PickCounterTypeString(Hdr->CounterType) );
        fprintf(fout, "CounterFreqHi       = %ld\n", Hdr->CounterFreqHi);
        fprintf(fout, "CounterFreq         = %lu\n", Hdr->CounterFreq);
        
        fprintf(fout, "Counter2Type        = %ld", Hdr->Counter2Type);
        fprintf(fout, PickCounterTypeString(Hdr->Counter2Type) );
        fprintf(fout, "Counter2FreqHi      = %ld\n", Hdr->Counter2FreqHi);
        fprintf(fout, "Counter2Freq        = %lu\n", Hdr->Counter2Freq);
        
        fprintf(fout, "LocationLatitude    = %lf\n", Hdr->LocationLatitude);
        fprintf(fout, "LocationLongitude   = %lf\n", Hdr->LocationLongitude);
        fprintf(fout, "LocationAltitude    = %lf\n", Hdr->LocationAltitude);
        fprintf(fout, "LocationSource      = %ld",   Hdr->LocationSource);
        fprintf(fout, PickSourceString(Hdr->LocationSource) );
        
        for ( i = 0 ; i < SRDAT_DIM_TIME ; i++ ) {
                fprintf(fout, "TimePoint[%d]:\n", i);
                
                fprintf(fout, "   Type             = %ld",   Hdr->TimePoint[i].Type);
                fprintf(fout, PickTimeTypeString(Hdr->TimePoint[i].Type) );
                fprintf(fout, "   SampleFromRun    = %ld\n", Hdr->TimePoint[i].SampleFromRun);
                fprintf(fout, "   SampleFromFile   = %ld\n", Hdr->TimePoint[i].SampleFromFile);
                fprintf(fout, "   GpsMarkIndex     = %ld\n", Hdr->TimePoint[i].GpsMarkIndex);
                
                fprintf(fout, "   Year             = %ld\n", Hdr->TimePoint[i].Year);
                fprintf(fout, "   Month            = %ld\n", Hdr->TimePoint[i].Month);
                fprintf(fout, "   Day              = %ld\n", Hdr->TimePoint[i].Day);
                fprintf(fout, "   YMDSource        = %ld",   Hdr->TimePoint[i].YMDSource);
                fprintf(fout, PickSourceString(Hdr->TimePoint[i].YMDSource) );
                
                fprintf(fout, "   Hour             = %ld\n", Hdr->TimePoint[i].Hour);
                fprintf(fout, "   Minute           = %ld\n", Hdr->TimePoint[i].Minute);
                fprintf(fout, "   Second           = %ld\n", Hdr->TimePoint[i].Second);
                fprintf(fout, "   Microsecond      = %ld\n", Hdr->TimePoint[i].Microsecond);
                fprintf(fout, "   HMSSource        = %ld",   Hdr->TimePoint[i].HMSSource);
                fprintf(fout, PickSourceString(Hdr->TimePoint[i].HMSSource) );
                
                fprintf(fout, "   Unused1          = %ld\n", Hdr->TimePoint[i].Unused1);
                }

        fprintf(fout, "LocationUnused1     = %ld\n", Hdr->LocationUnused1);
        fprintf(fout, "DigitalUnused1      = %ld\n", Hdr->DigitalUnused1);
        fprintf(fout, "TriggerUnused1      = %ld\n", Hdr->TriggerUnused1);

}


//------------------------------------------------------------------------------
// ROUTINE: PickTimeTypeString
// PURPOSE: Translate "time type" integer constant from header into a user 
//          readable string.  See srdat.h for constants.
//------------------------------------------------------------------------------
char *PickTimeTypeString( long Type ) {
        
        if (Type == SRDAT_TIME_TYPE_NONE)
                return( " (none)\n");

        else if (Type == SRDAT_TIME_TYPE_FIRSTPT)
                return( " (first point)\n");

        else if (Type == SRDAT_TIME_TYPE_LASTPT)
                return( " (last point)\n");

        else if (Type == SRDAT_TIME_TYPE_FIRSTGPS)
                return( " (first GPS point)\n");

        else if (Type == SRDAT_TIME_TYPE_LASTGPS)
                return( " (last GPS point)\n");

        else
                return( "\n");
}

//------------------------------------------------------------------------------
// ROUTINE: PickSourceString
// PURPOSE: Translate "source" integer constant from header into a user 
//          readable string.  See srdat.h for constants.
//------------------------------------------------------------------------------
char *PickSourceString( long Source ) {
        
        if (Source == SRDAT_SOURCE_NONE)
                return( " (none)\n");

        else if (Source == SRDAT_SOURCE_CALC)
                return( " (from calculation)\n");

        else if (Source == SRDAT_SOURCE_PC)
                return( " (from PC)\n");

        else if (Source == SRDAT_SOURCE_GPS)
                return( " (from GPS)\n");

        else if (Source == SRDAT_SOURCE_USER)
                return( " (from user)\n");

        else if (Source == SRDAT_SOURCE_ACQ)
                return( " (from acq program)\n");

        else if (Source == SRDAT_SOURCE_OBC)
                return( " (from GPS + on board counter)\n");

        else if (Source == SRDAT_SOURCE_EST)
                return( " (from GPS estimate)\n");

        else
                return( "\n");
}

//------------------------------------------------------------------------------
// ROUTINE: PickDataCodingString
// PURPOSE: Translate "data coding" integer constant from header into a user 
//          readable string.  See srdat.h for constants.
//------------------------------------------------------------------------------
char *PickDataCodingString( long DataCoding ) {
        
        if (DataCoding == SRDAT_DATACODING_SIGNED)
                return( " (signed)\n");

        else if (DataCoding == SRDAT_DATACODING_OFFSET)
                return( " (offset)\n");

        else
                return( "\n");
}


//------------------------------------------------------------------------------
// ROUTINE: PickCounterTypeString
// PURPOSE: Translate "counter type" integer constant from header into a user 
//          readable string.  See srdat.h for constants.
//------------------------------------------------------------------------------
char *PickCounterTypeString( long Type ) {
        
        if (Type == SRDAT_COUNTER_UNKNOWN)
                return( " (unknown)\n");

        else if (Type == SRDAT_COUNTER_INT64)
                return( " (64 bit integer)\n");

        else if (Type == SRDAT_COUNTER_HIGH32LOW32)
                return( " (2 long words, Linux timeval)\n");

        else if (Type == SRDAT_COUNTER_OTHER)
                return( " (other)\n");

        else
                return( "\n");
}

//------------------------------------------------------------------------------
// ROUTINE: PickAtodModelString
// PURPOSE: Translate "A/D model" integer constant from header into a user 
//          readable string.  See srdat.h for constants.
//------------------------------------------------------------------------------
char *PickAtodModelString( long AtodModel ) {
        
        if (AtodModel == SRDAT_ATODMODEL_UNKNOWN)
                return( " (unknown)\n");

        else if (AtodModel == SRDAT_ATODMODEL_PAR1CH)
                return( " (PAR1CH)\n");

        else if (AtodModel == SRDAT_ATODMODEL_PAR4CH)
                return( " (PAR4CH)\n");

        else if (AtodModel == SRDAT_ATODMODEL_PAR8CH)
                return( " (PAR8CH)\n");

        else if (AtodModel == SRDAT_ATODMODEL_USB1CH)
                return( " (USB1CH)\n");

        else if (AtodModel == SRDAT_ATODMODEL_USB4CH)
                return( " (USB4CH)\n");

        else if (AtodModel == SRDAT_ATODMODEL_USB8CH)
                return( " (USB8CH)\n");

        else
                return( "\n");
}

//------------------------------------------------------------------------------
// ROUTINE: PickGpsModelString
// PURPOSE: Translate "GPS model" integer constant from header into a user 
//          readable string.  See srdat.h for constants.
//------------------------------------------------------------------------------
char *PickGpsModelString( long GpsModel ) {
        
        if (GpsModel == SRDAT_GPSMODEL_UNKNOWN)
                return( " (unknown)\n");

        else if (GpsModel == SRDAT_GPSMODEL_NONE)
                return( " (none)\n");

        else if (GpsModel == SRDAT_GPSMODEL_TRIMBLE)
                return( " (TRIMBLE)\n");

        else if (GpsModel == SRDAT_GPSMODEL_GARMIN)
                return( " (GARMIN)\n");

        else if (GpsModel == SRDAT_GPSMODEL_ONCORE)
                return( " (Oncore GT+)\n");

        else if (GpsModel == SRDAT_GPSMODEL_PCTIME)
                return( " (PC System Time)\n");

        else if (GpsModel == SRDAT_GPSMODEL_MAX)
                return( " (unknown2)\n");

        else
                return( "\n");
}

//------------------------------------------------------------------------------
// ROUTINE: PickFileNameTypeString
// PURPOSE: Translate "file name type" integer constant from header into a user 
//          readable string.  See srdat.h for constants.
//------------------------------------------------------------------------------
char *PickFileNameTypeString( long FileNameType ) {
        
        if (FileNameType == SRDAT_FILENAMETYPE_UNKNOWN)
                return( " (Unknown)\n" );

        else if (FileNameType == SRDAT_FILENAMETYPE_NONE)
                return( " (None)\n" );

        else if (FileNameType == SRDAT_FILENAMETYPE_SINGLE)
                return( " (Single)\n" );

        else if (FileNameType == SRDAT_FILENAMETYPE_SEQDECIMAL)
                return( " (Sequential Decimal)\n" );

        else if (FileNameType == SRDAT_FILENAMETYPE_SEQHEX)
                return( " (Sequential Hex)\n" );

        else if (FileNameType == SRDAT_FILENAMETYPE_TIME)
                return( " (Time based)\n" );

        else
                return( "\n" );
}

//------------------------------------------------------------------------------
// ROUTINE: PickTriggerTypeString
// PURPOSE: Translate "trigger type" integer constant from header into a user 
//          readable string.  See srdat.h for constants.
//------------------------------------------------------------------------------
char *PickTriggerTypeString( long TriggerType ) {
        
        if (TriggerType == SRDAT_TRIGGER_UNKNOWN)
                return( " (Unknown)\n" );

        else if (TriggerType == SRDAT_TRIGGER_NONE)
                return( " (None)\n" );

        else if (TriggerType == SRDAT_TRIGGER_THRESHOLD)
                return( " (Amplitude Threshold)\n" );

        else if (TriggerType == SRDAT_TRIGGER_DIGLEVEL)
                return( " (Digital Level)\n" );

        else if (TriggerType == SRDAT_TRIGGER_DIGEDGE)
                return( " (Digital Edge)\n" );

        else
                return( "\n");
}

//------------------------------------------------------------------------------
// ROUTINE: PickTriggerParm1String
// PURPOSE: Translate trigger parameter 1 integer constant from header into a 
//          user readable string.  See srdat.h for constants.
//------------------------------------------------------------------------------
char *PickTriggerParm1String( long TriggerType, long TriggerParm1 ) {
        
        if (TriggerType == SRDAT_TRIGGER_THRESHOLD)
                return( " (Threshold)\n" );

        else if (TriggerType == SRDAT_TRIGGER_DIGLEVEL)
                return( " (Digital Input Line)\n" );

        else if (TriggerType == SRDAT_TRIGGER_DIGEDGE)
                return( " (Digital Input Line)\n" );

        else
                return( "\n" );
}


//------------------------------------------------------------------------------
// ROUTINE: PickTriggerParm2String
// PURPOSE: Translate trigger parameter 2 integer constant from header into a 
//          user readable string.  See srdat.h for constants.
//------------------------------------------------------------------------------
char *PickTriggerParm2String( long TriggerType, long TriggerParm2 ) {

        long channelBits;
        int  i, nchan;
 static char TmpStr[30];
        
        if (TriggerType == SRDAT_TRIGGER_THRESHOLD) {
                if ( TriggerParm2 == 0 )
                        return( " (No channels)\n" );
                else {
                        channelBits = TriggerParm2;
                        nchan       = 0;
                        for ( i = 0 ; i < 32 ; i++ ) {
                                if ( channelBits & 0x1 )
                                        nchan++;
                                channelBits /= 2;
                                }
                        sprintf( TmpStr, " = 0x%lX (%d channels)\n",
                                 TriggerParm2, nchan );
                        return( TmpStr );
                        }
                }

        else if (TriggerType == SRDAT_TRIGGER_DIGLEVEL) {
                if ( TriggerParm2 == SRDAT_TRIGLEVEL_LOW )
                        return( " (Low)\n" );
                else if ( TriggerParm2 == SRDAT_TRIGLEVEL_HIGH )
                        return( " (High)\n" );
                else
                        return( "\n" );
                }

        else if (TriggerType == SRDAT_TRIGGER_DIGEDGE) {
                if ( TriggerParm2 == SRDAT_TRIGEDGE_FALLING )
                        return( " (Falling)\n" );
                else if ( TriggerParm2 == SRDAT_TRIGEDGE_RISING )
                        return( " (Rising)\n" );
                else
                        return( "\n" );
                }

        else
                return( "\n");
}

//------------------------------------------------------------------------------
// ROUTINE: PickTimeMethodString
// PURPOSE: Translate "time method" integer constant from header into a user 
//          readable string.  See pargps.h or this file for constants.
//------------------------------------------------------------------------------
char *PickTimeMethodString( long TimeMethod ) {
        
        if (TimeMethod == TIME_METHOD_NONE)
                return( "Default time" );

        else if (TimeMethod == TIME_METHOD_SPS)
                return( "Sample rate time" );

        else if (TimeMethod == TIME_METHOD_GPS)
                return( "GPS time" );

        else if (TimeMethod == TIME_METHOD_OBC)
                return( "GPS+ time" );

        else if (TimeMethod == TIME_METHOD_PCT)
                return( "PC/NTP time" );

        else
                return( "" );
}

//------------------------------------------------------------------------------
// ROUTINE: ShowTag
// PURPOSE: Write record tag information to the ASCII output file.
//------------------------------------------------------------------------------
void ShowTag( FILE *fout, SRDAT_TAG *Tag ) {

        char cTagType[5];
        char *TypeStr;

        
        // Check tag size
        
        if (sizeof(*Tag) != SRDAT_TAGSIZE)
                ErrorWarning( "Tag size not right", 1 );

        strncpy( cTagType,((char *)(&Tag->TagId)), 4);
        cTagType[4] = '\0';


        // Set up tag type string

        switch (Tag->TagId) {
                case SRDAT_TAGID_ANALOG16:
                        TypeStr = "16 bit analog data";
                        break;
                case SRDAT_TAGID_ANALOG24:
                        TypeStr = "24 bit analog data";
                        break;
                case SRDAT_TAGID_GPSPPS:
                        TypeStr = "GPS pps data";
                        break;
                case SRDAT_TAGID_GPSPPST:
                        TypeStr = "GPS pps/time data";
                        break;
                case SRDAT_TAGID_GPSSERIAL:
                        TypeStr = "GPS serial data";
                        break;
                case SRDAT_TAGID_EOF:
                        TypeStr = "End of File";
                        break;
                default:
                        TypeStr = "Unknown";
                        break;
                }
        

        // Show tag info

        fprintf( fout, "\nRecord Type: %s,  Tag: %s,  Bytes: %lu\n",
                                             TypeStr, cTagType, Tag->Nbytes );
        
}


//------------------------------------------------------------------------------
// ROUTINE: ShowData
// PURPOSE: Write record data to the ASCII output file.  The format of the data
//          depends on the record type.
//------------------------------------------------------------------------------
void ShowData( FILE *fout, SRDAT_HDR *Hdr, void *Buffer,
               long DataType, unsigned long Nbytes,
               int HexFormat, int TimeFormat, int TimeMethod, 
               int ShowPcTime, int ShowAllInfo ) {

        char DataFmt16[MAXDATAFMT], DataFmt24[MAXDATAFMT], TimeFmt[MAXDATAFMT];
        char TimeStr[MAXTIMESTR];
        char cTagType[5], MarkStar;
        int i, j, Line;
        long ToShow, Type, DPdiff, DPTdiff;
        long *Analog24;
        short *Analog16;
        PPSDATA *GpsPps;
        PPSTDATA *GpsPpsPctime;
        SERIALDATA *GpsSerial;
        char *Unknown;
        double TmpDiff;
        unsigned long ii;
        

        // Show data

        Type = DataType;
        strncpy( cTagType,((char *)(&Type)), 4);
        cTagType[4] = '\0';

        if (HexFormat == 1) {
                strncpy( DataFmt16, "  0x%04X",  MAXDATAFMT-1 );
                strncpy( DataFmt24, "  0x%08lX", MAXDATAFMT-1 );
                }
        else {
                strncpy( DataFmt16, "  %+6d",   MAXDATAFMT-1 );
                strncpy( DataFmt24, "   %+9ld", MAXDATAFMT-1 );
                }
        DataFmt16[MAXDATAFMT-1] = '\0';
        DataFmt24[MAXDATAFMT-1] = '\0';

        if (TimeFormat == TIME_FORMAT_SEC)
                strncpy( TimeFmt, " %18s", MAXDATAFMT-1 );
        else // if (TimeFormat == TIME_FORMAT_YMD)
                strncpy( TimeFmt, " %26s", MAXDATAFMT-1 );
        TimeFmt[MAXDATAFMT-1] = '\0';
                

        
        switch (DataType) {
            case SRDAT_TAGID_ANALOG16:
                    Analog16 = (short*)Buffer;
                    ToShow = Nbytes;
                    j = 0;
                    if (ShowAllInfo) {
                            fprintf( fout, "Total Pt#   Pt#" );
                            for ( i = 0 ; i < Hdr->Channels ; i++ )
                                    fprintf( fout, "  %6s", Hdr->ChannelList[i].Title );
                            fprintf( fout, "\n" );
                            }
                    while (ToShow > 0) {
                            if (Hdr->ChannelMarkIndex == SRDAT_INDEXCHANNEL_NONE
                                || Analog16[Hdr->ChannelMarkIndex] == 0x0)
                                    MarkStar = MARKCHAR_NONE;
                            else
                                    MarkStar = MARKCHAR_FOUND;
                            if (ShowAllInfo)
                                    fprintf( fout, "%010ld  %03d", CurrentPt16, j );
                            for ( i = 0 ; i < Hdr->Channels ; i++ ) {
                                    fprintf( fout, DataFmt16, *Analog16++ );
                                    ToShow -= Hdr->AtodWordSize;
                                    }
                            if (ShowAllInfo)
                                    fprintf( fout, " %c%c\n", MarkStar, MarkStar );
                            else
                                    fprintf( fout, "\n" );
                            j++; CurrentPt16++;
                            }
                    break;
                        
            case SRDAT_TAGID_ANALOG24:
                    Analog24 = (long*)Buffer;
                    ToShow = Nbytes;
                    j = 0;
                    if (ShowAllInfo) {
                            fprintf( fout, "Total Pt#   Pt#" );
                            for ( i = 0 ; i < Hdr->Channels ; i++ )
                                    fprintf( fout, "  %10s", Hdr->ChannelList[i].Title );
                            if (TimeFormat != TIME_FORMAT_NONE) {
                                    fprintf( fout, TimeFmt, PickTimeMethodString(TimeMethod) );
                                    if (ShowPcTime && TimeMethod != TIME_METHOD_PCT )
                                            fprintf( fout, TimeFmt,
                                                     PickTimeMethodString(TIME_METHOD_PCT) );
                                    }
                            fprintf( fout, "\n" );
                            }
                    while (ToShow > 0) {
                            if (Hdr->ChannelMarkIndex == SRDAT_INDEXCHANNEL_NONE
                                || Analog24[Hdr->ChannelMarkIndex] == 0x0)
                                    MarkStar = MARKCHAR_NONE;
                            else
                                    MarkStar = MARKCHAR_FOUND;
                            if (ShowAllInfo)
                                    fprintf( fout, "%010ld  %03d", CurrentPt24, j );
                            for ( i = 0 ; i < Hdr->Channels ; i++ ) {
                                    fprintf( fout, DataFmt24, *Analog24++ );
                                    ToShow -= Hdr->AtodWordSize;
                                    }
                            if (TimeFormat != TIME_FORMAT_NONE) {
                                    ComputeTime( Hdr, TimeMethod, TimeFormat,
                                                 CurrentPt24, TimeStr );
                                    fprintf( fout, " %s", TimeStr );
                                    if (ShowPcTime &&
                                        TimeMethod != TIME_METHOD_PCT ) {
                                            ComputeTime( Hdr, TIME_METHOD_PCT,
                                                         TimeFormat, CurrentPt24,
                                                         TimeStr );
                                            fprintf( fout, " %s", TimeStr );
                                            }
                                    }
                            if (ShowAllInfo)
                                    fprintf( fout, " %c%c\n", MarkStar, MarkStar );
                            else
                                    fprintf( fout, "\n" );
                            j++; CurrentPt24++;
                            }
                    break;
                        
#ifdef GPS_AVAILABLE
            case SRDAT_TAGID_GPSPPS:
                    if (!ShowAllInfo)
                            break;
                    GpsPps = (PPSDATA*)Buffer;
                    ToShow = Nbytes;
                    while (ToShow > 0) {
                            ParGpsLargeIntDiffFull(
                                    GpsPps->CountAtDready,
                                    GpsPps->CountAtPps,
                                    NULL,
                                    &TmpDiff );
                            DPdiff = (long)TmpDiff;
                            fprintf(fout,"PpsEvent=%u, Cnt @Pps=0x%08lX%08lX @Dready=0x%08lX%08lX (D-P=%ld)\n",
                                      GpsPps->PpsEventNum,
                                      GpsPps->CountAtPps.u.HighPart,
                                      GpsPps->CountAtPps.u.LowPart,
                                      GpsPps->CountAtDready.u.HighPart,
                                      GpsPps->CountAtDready.u.LowPart,
                                      DPdiff);
                            GpsPps++;
                            ToShow -= sizeof(PPSDATA);
                            }
                    break;
                        
            case SRDAT_TAGID_GPSPPST:
                    if (!ShowAllInfo)
                            break;
                    GpsPpsPctime = (PPSTDATA*)Buffer;
                    ToShow = Nbytes;
                    while (ToShow > 0) {
                            ParGpsLargeIntDiffFull(
                                    GpsPpsPctime->CountAtDready,
                                    GpsPpsPctime->CountAtPps,
                                    NULL,
                                    &TmpDiff );
                            DPdiff = (long)TmpDiff;
                            ParGpsLargeIntDiffFull(
                                    GpsPpsPctime->PctimeAtDready,
                                    GpsPpsPctime->PctimeAtPps,
                                    NULL,
                                    &TmpDiff );
                            DPTdiff = (long)TmpDiff;
                            fprintf(fout,"PpsEvent=%u (0x%X)  Cnt  @Pps=0x%08lX%08lX @Dready=0x%08lX%08lX (D-P=%ld)\n",
                                      GpsPpsPctime->PpsEventNum,
                                      GpsPpsPctime->PpsEventNum,
                                      GpsPpsPctime->CountAtPps.u.HighPart,
                                      GpsPpsPctime->CountAtPps.u.LowPart,
                                      GpsPpsPctime->CountAtDready.u.HighPart,
                                      GpsPpsPctime->CountAtDready.u.LowPart,
                                      DPdiff);
                            fprintf(fout,"         %u (0x%X)  Time @Pps=0x%08lX%08lX @Dready=0x%08lX%08lX (D-P=%ld)\n",
                                      GpsPpsPctime->PpsEventNum,
                                      GpsPpsPctime->PpsEventNum,
                                      GpsPpsPctime->PctimeAtPps.u.HighPart,
                                      GpsPpsPctime->PctimeAtPps.u.LowPart,
                                      GpsPpsPctime->PctimeAtDready.u.HighPart,
                                      GpsPpsPctime->PctimeAtDready.u.LowPart,
                                      DPTdiff);
                            GpsPpsPctime++;
                            ToShow -= sizeof(PPSTDATA);
                            }
                    break;
                        
            case SRDAT_TAGID_GPSSERIAL:
                    if (!ShowAllInfo)
                            break;
                    GpsSerial = (SERIALDATA*)Buffer;
                    ToShow = Nbytes;
                    while (ToShow > 0) {
                            fprintf(fout, "Serial for PpsEvent = %u (0x%X), NMEA string count = %d\n",
                                      GpsSerial->PpsEventNum,
                                      GpsSerial->PpsEventNum,
                                      GpsSerial->NmeaCount);
                            for ( i = 0 ; i < GpsSerial->NmeaCount ; i++ ) {
                                    fprintf(fout,"%s",
                                       &GpsSerial->NmeaMsg[i*MAX_NMEA_SIZE]);
                                    }
                            GpsSerial++;
                            ToShow -= sizeof(SERIALDATA);
                            }
                    break;
#else                    
            case SRDAT_TAGID_GPSPPS:
            case SRDAT_TAGID_GPSPPST:
            case SRDAT_TAGID_GPSSERIAL:
                    if (!ShowAllInfo)
                            break;
                    TmpDiff      = (double)0.0;
                    DPdiff       = 0;
                    DPTdiff      = 0;
                    GpsPps       = NULL;
                    GpsPpsPctime = NULL;
                    GpsSerial    = NULL;
                    fprintf( fout, "Compile with GPS_AVAILABLE to display PPS and serial records\n" );
                    printf("DAT2ASC must be compiled with GPS_AVAILABLE in\n" );
                    printf("order to understand GPS PPS and serial records.\n" );
                    // fall through to default case
#endif

            default:
                    if (!ShowAllInfo)
                            break;
                    fprintf(fout,"Unknown data type 0x%lX\n", DataType );

                    Line = 0;
                    Unknown = (char *)Buffer;

                    for ( ii = 0 ; ii < Nbytes ; ii++ ) {
                            fprintf( fout, "%02X ", (Unknown[ii] & 0xFF) );
                            Line++;
                            if (Line > 9) {
                                    fprintf( fout, "\n" );
                                    Line = 0;
                                    }
                            }
                    fprintf( fout, "\n" );
                        
                    break;

            } // end switch
}


//------------------------------------------------------------------------------
// ROUTINE: ReadDatRecord
// PURPOSE: Read Analog, Pps, or Serial type records in from the binary .dat
//          data file.
//------------------------------------------------------------------------------
int ReadDatRecord( SRDAT_HDR *Hdr,
                   void  *PC_BUFFER, unsigned int *npts,
                   void *PPS_BUFFER, unsigned int *npps,
                   void *SER_BUFFER, unsigned int *nser
                 ) {

        SRDAT_TAG    Tag;
        unsigned int nbytes;


        // Initialize values

        *npts = *npps = *nser = 0;



        // Read data record tag to determine data type
        
        if ( !SrDatTagRead( Fin, &Tag ) ) {
                ErrorNotice( "Could not read .DAT tag (ReadDatRecord)" );
                return( 0 );
                }


        // Check data size (only matters in DOS version)

        nbytes = (unsigned int)Tag.Nbytes;
        if ((unsigned long)nbytes != Tag.Nbytes)
                ErrorWarning( "Error data record exceeds size allowed by unsigned int.", 1 );



        // Read in analog, pps, or serial data buffer

        if (Tag.TagId == SRDAT_TAGID_ANALOG24) {

                if ( !SrDatDataReadAnalog24( Fin, PC_BUFFER, nbytes, npts ) )
                        ErrorWarning( "Could not read .DAT analog data", 1 );
                }


        else if (Tag.TagId == SRDAT_TAGID_GPSPPS) {

                if ( !SrDatDataReadGpsPps( Fin, PPS_BUFFER, nbytes, npps,
                                           (int)Hdr->CounterType ) )
                        ErrorWarning( "Could not read .DAT PPS data", 1 );
                }


        else if (Tag.TagId == SRDAT_TAGID_GPSPPST) {

                if ( !SrDatDataReadGpsPpsT( Fin, PPS_BUFFER, nbytes, npps ) )
                        ErrorWarning( "Could not read .DAT PPS(T) data", 1 );
                }


        else if (Tag.TagId == SRDAT_TAGID_GPSSERIAL) {

                if ( !SrDatDataReadGpsSerial( Fin, SER_BUFFER, nbytes, nser ) )
                        ErrorWarning( "Could not read .DAT serial data", 1 );
                }


        else if (Tag.TagId == SRDAT_TAGID_EOF) {
                return( 0 );
                } // end TAGID_EOF

        else {
                printf( "Unexpected tag TagId found 0x%lX\n", Tag.TagId);
                }



        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: SelectTimeMethod
// PURPOSE: Select the best method for computing the time for each data point
//          based on header info using the following header fields:
//          DatGpsModel, ChannelObcIndex, ChannelMarkIndex
//------------------------------------------------------------------------------
int SelectTimeMethod( SRDAT_HDR *Hdr ) {
        
        int Method;


        // Error check

        if ( !Hdr )
                Method = TIME_METHOD_SPS;


        // If PC time is specified for the GPS model, use it

        else if ( Hdr->DatGpsModel == SRDAT_GPSMODEL_PCTIME )
                Method = TIME_METHOD_PCT;

        
        // Or if there is an OBC channel, use it

        else if ( Hdr->ChannelObcIndex != SRDAT_INDEXCHANNEL_NONE )
                Method = TIME_METHOD_OBC;

        
        // Or if there is GPS info, use it
        
        else if ( Hdr->ChannelMarkIndex != SRDAT_INDEXCHANNEL_NONE )
                Method = TIME_METHOD_GPS;

        
        // Otherwise, just use sps

        else
                Method = TIME_METHOD_SPS;


        return( Method );
}


//------------------------------------------------------------------------------
// ROUTINE: FillTimeStr
// PURPOSE: Fill a string with time information in either Y/M/D H:M:S or
//          seconds since 1970 format as requested.
//------------------------------------------------------------------------------
int FillTimeStr( int TimeFormat, double Time, char *TimeStr ) {

        int  Year, Month, Day, Hour, Minute, Second;
        long Microsecond;

        // Error check

        if ( !TimeStr )
                return( 0 );
        
        
        // Fill time string with selected format

        if ( TimeFormat == TIME_FORMAT_YMD ) {
                
                ParGpsSecTimeSplit( Time, &Year, &Month, &Day, 
                                    &Hour, &Minute, &Second, &Microsecond );
        
                sprintf( TimeStr, "%4d/%02d/%02d %02d:%02d:%02d.%06ld",
                         Year, Month, Day, Hour, Minute, Second, Microsecond );
                }
        
        else if ( TimeFormat == TIME_FORMAT_SEC )
                sprintf( TimeStr, "%18.6lf", Time );

        else
                return( 0 );

        
        return( 1 );

}


//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Print program usage message to the screen.
//------------------------------------------------------------------------------
void Usage( char *msg ) {

        int i;
        
        // Select error line depending on 
        // 1) no message, 2) help requested, or 3) message given

        if ( !msg )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( msg, "?" ) == 0 )
                printf( "\n" );

        else
                printf( "\nERROR: Command line usage (%s).\n\n", msg );


        for ( i = 0 ; UsageMsg[i] ; i++ )
                printf( "%s\n", UsageMsg[i] );

        exit( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ErrorWarning
// PURPOSE: Print the supplied error message to the screen and exit cleanly
//          with the supplied status.
//------------------------------------------------------------------------------
void ErrorWarning( char *msg, int status ) {
        ErrorNotice( msg );
        CleanExit( status );
}


//------------------------------------------------------------------------------
// ROUTINE: ErrorNotice
// PURPOSE: Print the supplied error message to the screen.
//------------------------------------------------------------------------------
void ErrorNotice( char *msg ) {
        printf("DAT2ASC ERROR: %s\n", msg );
}


//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Exit cleanly by ensuring that all open files are closed before exit.
//------------------------------------------------------------------------------
void CleanExit( int status ) {

        // Close any open files

        if ( Fin != NULL )
                fclose( Fin );

        if ( Fout != NULL )
                fclose( Fout );

        if ( Fts != NULL )
                fclose( Fts );

        // Exit

        exit( status );
}


//------------------------------------------------------------------------------
// The behavior of the following three functions is very different
// depending on whether or not GPS info is available.

#ifdef GPS_AVAILABLE

//------------------------------------------------------------------------------
// ROUTINE: InitTimeStamps (GPS VERSION)
// PURPOSE: Initialize time stamp information that will be used to determine the
//          time of each sample point.  This is done by reading through all the
//          input .dat files gathering up the Analog, Pps, and Serial information
//          associated with each GPS PPS signal and associating them all together
//          into a time stamp structure.  This time stamp info is saved in a
//          temporary file to remove any hardwired limits on how many files can
//          be processes at one time.
//
int InitTimeStamps( void ) {

        unsigned int PrevEventNum, PpsEvent, Npts, Npps, Nser;
        int          TsTitle, PrevUnset, ok, TsValid, i, GpsInvalidWarning;
        int          NotDone, Nvalid, SerialDelay, GpsModel;
        long         NumChannels, NumSamples, Counter2Type, CounterType;
        long         StartSample, MarkChannel, ObcChannel, SeqNumSave;
        double       PrevPpsSecTime, TpSecTime, LastSecSince1970, LastDiff;
        TIMESTAMP    TSx;
        TIMEPT       Tp;
        SR64BIT      Zero64;
        long         LastSample[2], LastCountObc[2];
        SR64BIT      LastCountAtPps[2], LastCountAtDready[2];
        SR64BIT      LastPctimeAtPps[2], LastPctimeAtDready[2];


        // Go through all .dat files initializing timestamps


        // Open timestamp file

        Fts = fopen( "TIMESTMP.TMP", "w+" );
        if (!Fts)
                ErrorWarning( "Can not open temporary file TIMESTMP.TMP", 1 );

        TsTitle = 1; // Request that we write a title the first time through
                
        
        
        // Make sure first file exists

        if ( !SrFileExist( FileIn ) ) {
                sprintf( Msg, "Can not open input file %s.", FileIn );
                ErrorWarning( Msg, 1 );
                }

        
        PrevUnset          = 1;
        SeqNumSave         = SeqNum;
        PrevEventNum       = INVALID_PPSEVENT;
        SerialDelay        = 0;
        Nvalid             = 0;
        TimeStamp[0].Valid = TS_VALID_NONE;
        TimeStamp[1].Valid = TS_VALID_NONE;
        Zero64.u.HighPart  = 0L;
        Zero64.u.LowPart   = 0L;
        TsValid            = TS_VALID_ALL;  // This is a global variable


        // Initialize ParGps library timestamp functions

        ParGpsTsInit();
                


        // Loop for several files

        printf( "Initializing TimeStamps:\n" );

        while ( 1 ) {

                // First check for existance of input file

                if ( !SrFileExist( FileIn ) ) {
                        printf( "No more files to process.  Stopping at %s.\n", FileIn );
                        break;
                        }
                
                printf( "Reading file %s ...\n", FileIn );

                
                // Open input file

                Fin = fopen( FileIn, "rb" );
                if (!Fin) {
                        printf( "ERROR: Can not open input file %s\n", FileIn );
                        break;
                        }


                // Read header, gather useful fields and check them

                if ( !SrDatHdrRead( Fin, &Hdr ) )
                        ErrorWarning( "Could not read .DAT header", 1 );

                if ( !SrDatHdrValidate( &Hdr ) )
                        ErrorWarning( "This is not a valid .DAT file", 1 );

                GpsModel             = Hdr.DatGpsModel;
                NumChannels          = Hdr.Channels;
                StartSample          = Hdr.FileStartPtNum;
                MarkChannel          = Hdr.ChannelMarkIndex;
                ObcChannel           = Hdr.ChannelObcIndex;
                ObcCountsPerSample   = (OBC_FREQ / Hdr.AtodSamplingRate + 0.5L);
                BestCountsBetweenPps = (unsigned long)Hdr.CounterFreq;
                CountsPerSecond      = (unsigned long)Hdr.Counter2Freq;
                CounterType          = Hdr.CounterType;
                Counter2Type         = Hdr.Counter2Type;
                TimeMethod           = SelectTimeMethod( &Hdr );
                GpsInvalidWarning    = 0;


                // When using PC time, we ignore NMEA serial data

                if (GpsModel == SRDAT_GPSMODEL_PCTIME)
                        TsValid = TS_VALID_MOST;



                // Get PC time in seconds from the header

                Tp = Hdr.TimePoint[0];
                ParGpsSecTimeCombine(
                                (int)Tp.Year, (int)Tp.Month, (int)Tp.Day,
                                (int)Tp.Hour, (int)Tp.Minute, (int)Tp.Second,
                                 Tp.Microsecond, &TpSecTime );


                // If the SPS method was selected, there won't be any
                // time info later in the file so make and save a
                // timestamp from the header info.
                        
                if ( TimeMethod == TIME_METHOD_SPS ) {
                        TSx.Valid          = TS_VALID_ALL;
                        TSx.PpsEvent       = INVALID_PPSEVENT;
                        TSx.Sample         = Hdr.FileStartPtNum;
                        TSx.CountObc       = 0;
                        TSx.CountAtPps     = Zero64;
                        TSx.CountAtDready  = Zero64;
                        TSx.SecSince1970   = TpSecTime;
                        TSx.NumSat         = 0;
                        TSx.YmdSource      = Tp.YMDSource;
                        TSx.HmsSource      = Tp.HMSSource;
                        TSx.PctimeAtPps    = Zero64;
                        TSx.PctimeAtDready = Zero64;
                        ParGpsTsFileWrite( Fts, &TSx, TsTitle );
                        TsTitle = 0;
                        }
                

                // In case there is no GPS YMD info available, this
                // PC time from header will start things off right

                if ( PrevUnset ) {
                        PrevPpsSecTime = TpSecTime;
                        PrevUnset      = 0;
                        }



                // Loop through processing entire file

                NotDone = 1;

                while ( NotDone ) {

                        // Read next data record

                        NotDone = ReadDatRecord( &Hdr,
                                                  PC_BUFFER, &Npts,
                                                 PPS_BUFFER, &Npps,
                                                 SER_BUFFER, &Nser
                                               );



                        // If it's one of the three desired types,
                        // process it into time stamps

                        if (Npts > 0)  {
                                NumSamples = Npts/NumChannels;
                                ParGpsTsProcessAnalog( PC_BUFFER, NumSamples, NumChannels,
                                                       StartSample, MarkChannel, 
                                                       ObcChannel, ObcCountsPerSample,
                                                       TsValid,
                                                       &PpsEvent, &Nvalid );
                                StartSample += NumSamples;
                                }

                        else if (Npps > 0) {
                                ParGpsTsProcessPps( PPS_BUFFER, Npps,
                                                    CountsPerSecond,
                                                    GpsModel, TsValid,
                                                    &PpsEvent, &Nvalid );
                                }


                        else if (Nser > 0) {
                                ParGpsTsProcessSerial( SER_BUFFER, Nser, 
                                                       SerialDelay, TsValid,
                                                       &PrevEventNum, &PrevPpsSecTime,
                                                       &PpsEvent, &Nvalid );
                                }


                        // Read the valid time stamps from the circular buffer
                        // maintained by the PARGPS library, correct the HMS source
                        // if using OBC to compute time, and write the time stamps
                        // out to the temporary file.

                        while ( Nvalid > 0 ) {
                                ParGpsTsReadValid( &TSx, TsValid );

                                if ( TSx.HmsSource == TIME_SOURCE_EST ||
                                     TSx.YmdSource == TIME_SOURCE_EST ) {
                                        if (GpsInvalidWarning == 0) {
                                                printf( "*** WARNING: Some GPS time computations used data without satellite lock\n" );
                                                GpsInvalidWarning = 1;
                                                }
                                        }
                                        
                                if ( TimeMethod == TIME_METHOD_OBC &&
                                     TSx.HmsSource == TIME_SOURCE_GPS )
                                        TSx.HmsSource = TIME_SOURCE_OBC;
                                if ( TimeMethod == TIME_METHOD_OBC &&
                                     TSx.HmsSource == TIME_SOURCE_EST )
                                        TSx.HmsSource = TIME_SOURCE_OBC;
                                
                                if ( TimeMethod == TIME_METHOD_PCT &&
                                     TSx.HmsSource == TIME_SOURCE_GPS )
                                        TSx.HmsSource = TIME_SOURCE_PC;
                                if ( TimeMethod == TIME_METHOD_PCT &&
                                     TSx.HmsSource == TIME_SOURCE_EST )
                                        TSx.HmsSource = TIME_SOURCE_PC;
                                
                                ParGpsTsFileWrite( Fts, &TSx, TsTitle );
                                TsTitle = 0;
                                Nvalid--;
                                // WCT: Maybe recompute best counts between pps here
                                }


                        }  // end of while NotDone


                fclose( Fin );   Fin = NULL;



                // If only a single file was requested, we're done
                
                if (!Sequential)
                        break;

                // Otherwise, compute the next sequential file name
                
                else {
                        SeqNum++;
                        if ( SeqType == SEQTYPE_HEX )
                                sprintf( FileIn,  "%08lX.DAT", SeqNum );
                        else // ( SeqType == SEQTYPE_DEC )
                                sprintf( FileIn,  "%08ld.DAT", SeqNum );
                        }

                } // end of while 1


        // We've gotten to the end of all files and have all valid timestamps
        // Now let's get any that are just missing the serial strings or
        // either the pps or analog record.

        LastSecSince1970              = TSx.SecSince1970;
        for ( i = 0 ; i < 2 ; i++ ) {
                LastSample[i]         = TSx.Sample;
                LastCountObc[i]       = TSx.CountObc;
                LastCountAtPps[i]     = TSx.CountAtPps;
                LastCountAtDready[i]  = TSx.CountAtDready;
                LastPctimeAtPps[i]    = TSx.PctimeAtPps;
                LastPctimeAtDready[i] = TSx.PctimeAtDready;
                }
        
        while ( TsValid = ParGpsTsReadNext( &TSx ) ) {
                

                // Serial info not valid, so replace it with an estimate

                if ( !(TsValid & TS_VALID_SERIAL) ) {
                        TSx.YmdSource    = TIME_SOURCE_CALC;
                        TSx.HmsSource    = TIME_SOURCE_CALC;
                        TSx.SecSince1970 = LastSecSince1970+1;
                        }


                // Pps info not valid, so replace it with an estimate

                if ( !(TsValid & TS_VALID_PPS) ) {
                        ParGpsLargeIntDiffFull( LastCountAtPps[0], LastCountAtPps[1],
                                                NULL, &LastDiff );
                        TSx.CountAtPps    = ParGpsLargeIntAdd( LastCountAtPps[0],
                                                (long)LastDiff );
                        ParGpsLargeIntDiffFull( LastCountAtDready[0], LastCountAtDready[1],
                                                NULL, &LastDiff );
                        TSx.CountAtDready = ParGpsLargeIntAdd( LastCountAtDready[0],
                                                (long)LastDiff );
                        
                        ParGpsLargeIntDiffFull( LastPctimeAtPps[0], LastPctimeAtPps[1],
                                                NULL, &LastDiff );
                        TSx.PctimeAtPps    = ParGpsLargeIntAdd( LastPctimeAtPps[0],
                                                (long)LastDiff );
                        ParGpsLargeIntDiffFull( LastPctimeAtDready[0], LastPctimeAtDready[1],
                                                NULL, &LastDiff );
                        TSx.PctimeAtDready = ParGpsLargeIntAdd( LastPctimeAtDready[0],
                                                (long)LastDiff );
                        }

                // Analog/OBC info not valid, so replace it with an estimate

                if ( !(TsValid & TS_VALID_ANALOG) ) {
                        TSx.Sample   = LastSample[0] +
                                      (LastSample[0] - LastSample[1]);
                        TSx.CountObc = LastCountObc[0] + 
                                      (LastCountObc[0] - LastCountObc[1]);
                        }
                
                ParGpsTsFileWrite( Fts, &TSx, 0 );
                
                LastSecSince1970     = TSx.SecSince1970;
                
                LastSample[1]        = LastSample[0];
                LastCountObc[1]      = LastCountObc[0];
                LastCountAtPps[1]    = LastCountAtPps[0];
                LastCountAtDready[1] = LastCountAtDready[0];
                
                LastSample[0]        = TSx.Sample;
                LastCountObc[0]      = TSx.CountObc;
                LastCountAtPps[0]    = TSx.CountAtPps;
                LastCountAtDready[0] = TSx.CountAtDready;
                
                } // end while TsValid

        
        // Restore orignal sequence number and input file name

        SeqNum = SeqNumSave;
        if (Sequential) {
                if ( SeqType == SEQTYPE_HEX )
                        sprintf( FileIn,  "%08lX.DAT", SeqNum );
                else // ( SeqType == SEQTYPE_DEC )
                        sprintf( FileIn,  "%08ld.DAT", SeqNum );
                }



        // Close the time stamp file

        fclose( Fts );  Fts = NULL;


        
        // Re-open time stamp file for reading

        Fts = fopen( "TIMESTMP.TMP", "r" );
        if (!Fts)
                ErrorWarning( "Can not open temporary file TIMESTMP.TMP", 1 );

        fgets( Msg, MAXFILENAME, Fts ); // Read past title


        // Get some timestamps to start us off
        // (get two except for SPS method which only needs one).
        
        ok = ParGpsTsFileRead( Fts, &TimeStamp[0], 0 );
        if (!ok) {
                printf( "Could not read time stamp 0\n" );
                return( 0 );
                }
        
        if ( TimeMethod == TIME_METHOD_SPS ) {
                // Fill in a dummy second timestamp
                TimeStamp[1]        = TimeStamp[0];
                TimeStamp[1].Sample = 0xFFFFFFFF;
                }
        else {
                ok = ParGpsTsFileRead( Fts, &TimeStamp[1], 0 );
                if (!ok) {
                        printf( "Could not read time stamp 1\n" );
                        return( 0 );
                        }
                }


        // When requesting GPS time, do a check against counter frequency

        if (TimeMethod == TIME_METHOD_GPS && Hdr.CounterFreqHi != 0L) {
                printf( "WARNING: PC counter frequency too high for this program to handle.\n" );
                printf( "         Please check for an updated version or contact SR for assistance.\n" );
                return( 0 );
                }


        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ComputeTime (GPS VERSION)
// PURPOSE: Use information in the previously constructed time stamp structures
//          to determine time of the specified sample point.
//
int ComputeTime( SRDAT_HDR *Hdr, int TimeMethod, int TimeFormat, long Sample, char *TimeStr ) {

        double Time;
        int    ok, TsValidLax;


        // Uses globals: TimeStamp{}, TsValid
        //               ObcCountsPerSample, BestCountsBetweenPps


        // Do extensive time computations based on timestamps prepared
        // for each PPS event.


        // Error check

        if (!Hdr || !TimeStr || TimeFormat == TIME_FORMAT_NONE )
                return( 0 );


        // Allow timestamps without serial info

        if (TsValid == TS_VALID_ALL)
                TsValidLax = TS_VALID_MOST;
        else
                TsValidLax = TsValid;
                


        // Update timestamps if current sample number is
        // beyond the right-most timestamp sample number

        if ( Sample > TimeStamp[1].Sample )
                UpdateTimeStamps( Sample );


        // Compute the time using the selected method

        if ( TimeMethod == TIME_METHOD_OBC )
                ok = ParGpsTsComputeTimeObc( &TimeStamp[0], &TimeStamp[1],
                                             ObcCountsPerSample,
                                             TsValidLax, Sample, &Time );

        else if ( TimeMethod == TIME_METHOD_GPS )
                ok = ParGpsTsComputeTimePps( &TimeStamp[0], &TimeStamp[1],
                                             BestCountsBetweenPps,
                                             TsValidLax, Sample, &Time );

        else if ( TimeMethod == TIME_METHOD_PCT ) {
                ok = ParGpsTsComputeTimePctime( &TimeStamp[0], &TimeStamp[1],
                                             Hdr->Counter2Freq, 
                                             TsValidLax, Sample, &Time );
                }

        else // ( TimeMethod == TIME_METHOD_SPS )
                ok = ParGpsTsComputeTimeSps( &TimeStamp[0], &TimeStamp[1],
                                             Hdr->AtodSamplingRate,
                                             TsValidLax, Sample, &Time );

        if (!ok) {
                strcpy( TimeStr, "Time error" );
                return( 0 );
                }


        // Fill in the time string to output

        FillTimeStr( TimeFormat, Time, TimeStr );

        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: UpdateTimeStamps (GPS VERSION)
// PURPOSE: Update the pair of time stamp structures currently in use to ensure
//          that they surround the specified sample point.
//
int UpdateTimeStamps( long Sample ) {

        TIMESTAMP TSx;
        // Uses globals: TimeMethod, TimeStamp[], Fts


        // No update required for SPS method

        if ( TimeMethod == TIME_METHOD_SPS )
                return( 1 );
                     
        
        while ( Sample > TimeStamp[1].Sample ) {

                if ( ParGpsTsFileRead( Fts, &TSx, 0 ) ) {
                        TimeStamp[0] = TimeStamp[1];
                        TimeStamp[1] = TSx;
                        }
                else
                        return( 0 );
                
                } // end while
        
        return( 1 );
}



#else // not GPS_AVAILABLE

//------------------------------------------------------------------------------
// ROUTINE: InitTimeStamps (GPS NOT AVAILABLE)
// PURPOSE: Initialize time stamp information that will be used to determine the
//          time of each sample point.  This is done by reading in the sampling
//          rate given in the header part of the first .dat file and saving this
//          info in a time stamp structure.
//
int InitTimeStamps( void ) {

        TIMEPT Tp;
        double TpSecTime;


        // Set global TimeMethod to NONE

        TimeMethod = TIME_METHOD_NONE;



        // Read header from first file

        printf( "Initializing TimeStamps:\n" );

        TimeStamp[0].Sample       = 0L;
        TimeStamp[0].SecSince1970 = (double)0.0;




        // Open input file

        Fin = fopen( FileIn, "rb" );
        if (!Fin) {
                sprintf( Msg, "Can not open input file %s", FileIn );
                ErrorWarning( Msg, 1 );
                }


        // Read header, gather useful fields and check them

        if ( !SrDatHdrRead( Fin, &Hdr ) )
                ErrorWarning( "Could not read .DAT header", 1 );

        if ( !SrDatHdrValidate( &Hdr ) )
                ErrorWarning( "This is not a valid .DAT file", 1 );

        fclose( Fin );   Fin = NULL;

        

        // Get PC time in seconds from the header and save time
        // as a "timestamp" (cutdown version since no pargps.h)

        Tp = Hdr.TimePoint[0];

        if (Tp.Type != SRDAT_TIME_TYPE_NONE) {
                ParGpsSecTimeCombine(
                                (int)Tp.Year, (int)Tp.Month,  (int)Tp.Day,
                                (int)Tp.Hour, (int)Tp.Minute, (int)Tp.Second,
                                 Tp.Microsecond, &TpSecTime );
        
                TimeStamp[0].Sample       = Tp.SampleFromRun;
                TimeStamp[0].SecSince1970 = TpSecTime;
                }
                

        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ComputeTime (GPS NOT AVAILABLE)
// PURPOSE: Use information in the previously constructed time stamp structure
//          to determine time of the specified sample point.
//
int ComputeTime( SRDAT_HDR *Hdr, int TimeMethod, int TimeFormat,
                 long Sample, char *TimeStr ) {

        double  Time;


        // Use PC time and sampling rate from the header to compute time.

        // Error check

        if ( !Hdr                           ||
             !TimeStr                       ||
             TimeFormat == TIME_FORMAT_NONE ||
             TimeMethod != TIME_METHOD_NONE )
                return( 0 );



        // Get the first time point from the header and use it plus the
        // sampling rate to determine the time.  Note that we checked
        // AtodSamplingRate in main to ensure it is valid so we won't
        // divide by zero here.

        Time = TimeStamp[0].SecSince1970 +
               (double)(Sample - TimeStamp[0].Sample) /
               Hdr->AtodSamplingRate;  



        // Fill in the time string to output

        FillTimeStr( TimeFormat, Time, TimeStr );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: UpdateTimeStamps (GPS NOT AVAILABLE)
// PURPOSE: Only one time stamp is used when there is no GPS info, so this
//          function is merely a place holder.
//
int UpdateTimeStamps( long Sample ) {

        return( 1 );
}




//------------------------------------------------------------------------------
// PARGPS LIBRARY REPLACEMENT HELPER FUNCTIONS
//
// The following functions are duplicates of those in the PARGPS
// library, but they are so handy we're repeating them here if the
// PARGPS library is not linked in.


#define USPERSEC  1000000L // microseconds / second

//------------------------------------------------------------------------------
// ROUTINE: ParGpsCalcTimeZoneCorr (GPS NOT AVAILABLE)
// PURPOSE: Calculate the time zone correction.
//
int ParGpsCalcTimeZoneCorr( long StartTime ) {
        
        time_t    TargetTime, NewTime;
        struct tm tm;
        int       ZoneCorrection;

        // This function does two conversions between time stored as a structure
        // and time as seconds since the epoch (1970-01-01).  They should cancel
        // out, but the first is done using GMT time and the second is done using
        // local time.  The difference gives us the amount of the local correction.
        // Subtract this value from a local time to get a GMT time.
        
        
        // Convert epoch seconds to time structure for gmt
        
        TargetTime = (time_t)StartTime;
        tm = *gmtime( &TargetTime );

        
        // Convert time structure for local to epoch seconds (force DST unknown)
        
        tm.tm_isdst = -1;
        NewTime = mktime( &tm );


        // Difference is local correction

        ZoneCorrection = (int)(NewTime - TargetTime);
        return( ZoneCorrection );
}


//------------------------------------------------------------------------------
// ROUTINE: ParGpsSecTimeSplit (GPS NOT AVAILABLE)
// PURPOSE: Split a double containing seconds since 1970 into its component
//          Y/M/D H:M:S.u fields
//
int ParGpsSecTimeSplit( double Time,
                        int *Year, int *Month, int *Day,
                        int *Hour, int *Minute, int *Second, 
                        long *Microsecond ) {

        time_t    ttTime;
        struct tm tmTime;
        double    usec;


        // Split time into pieces (for UTC)

        ttTime = (time_t)Time;
        tmTime = *gmtime( &ttTime );
        usec   = Time - (double)ttTime + 0.0000005;
        
        // Save pieces
        
        if (Year)         *Year        = tmTime.tm_year + 1900;
        if (Month)        *Month       = tmTime.tm_mon + 1;
        if (Day)          *Day         = tmTime.tm_mday;
        if (Hour)         *Hour        = tmTime.tm_hour;
        if (Minute)       *Minute      = tmTime.tm_min;
        if (Second)       *Second      = tmTime.tm_sec;
        if (Microsecond)  *Microsecond = (long)(usec * USPERSEC);

        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ParGpsSecTimeCombine (GPS NOT AVAILABLE)
// PURPOSE: Combine the component Year/Month/Day Hour:Minute:Second.MicroSecond
//          fields into a double containing seconds since 1970.
//          
//
int ParGpsSecTimeCombine( int Year, int Month, int Day,
                          int Hour, int Minute, int Second, 
                          long MicroSecond, double *Time ) {

        time_t    ttTime;
        struct tm tmTime;

        // Set default YMD values when 0's are passed in

        if (Year == 0)   Year  = 1970;
        if (Month == 0)  Month = 1;
        if (Day == 0)    Day   = 1;


        // Fill pieces
        
        tmTime.tm_year  = Year - 1900;
        tmTime.tm_mon   = Month - 1;
        tmTime.tm_mday  = Day;
        tmTime.tm_hour  = Hour;
        tmTime.tm_min   = Minute;
        tmTime.tm_sec   = Second;
        tmTime.tm_wday  = 0;
        tmTime.tm_yday  = 0;
        tmTime.tm_isdst = -1;


        // Combine time from GMT pieces to epoch seconds using local correction,
        // then use the zone offset to remove the local correction

        ttTime = mktime( &tmTime );
        if (ttTime == (time_t)-1)
                return( 0 );
        ttTime -= ParGpsCalcTimeZoneCorr( ttTime );
        if (Time)  *Time  = (double)ttTime + (double)MicroSecond / (double)USPERSEC;
        
        return( 1 );
}

#endif // GPS_AVAILABLE
