/* FILE: srhelper.c             Copyright (c), Symmetric Research, 2002-2006

These are the SR user library helper functions.  They are mostly
wrappers around some utility functions that differ between various OS's.

*/

#include <stdio.h>       // fileno function, etc
#include <sys/stat.h>    // fstat function  (VC++ can use / in path)


// OS dependent includes ...

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <windows.h>     // windef.h -> winnt.h -> defines GENERIC_READ etc
#include <io.h>          // access function
#include <conio.h>       // kbhit, getch functions
#define OsAccess _access
#define OsFileno _fileno


#elif defined( SROS_WIN9X )
#include <windows.h>     // windef.h -> winnt.h -> defines GENERIC_READ etc
#include <stdlib.h>      // sleep function now from windows.h
#include <io.h>          // access function
#include <conio.h>       // kbhit, getch functions
#define OsAccess _access
#define OsFileno _fileno


#elif defined( SROS_MSDOS )
#include <io.h>          // access function
#include <conio.h>       // kbhit, getch functions
#include <sys\timeb.h>   // ftime function
#define OsAccess _access
#define OsFileno _fileno


#elif defined( SROS_LINUX )
#include <unistd.h>      // access function and STDIN_FILENO
#include <fcntl.h>       // O_xxx macros
#define OsAccess access
#define OsFileno fileno


#endif  // SROS_xxxxx


#if !defined( SRLOCAL )
#define SRLOCAL static
#endif


#include "srhelper.h"



// Prototypes for OS dependent helper functions

SRLOCAL int  OsSetNonBlock( int value );
SRLOCAL int  OsGetKey( int wait );
SRLOCAL void OsFlushPrint( void );
SRLOCAL int  OsSleep( int ms, int hardwait );




//------------------------------------------------------------------------------
// ROUTINE: SrHelperRev
// PURPOSE: Allows the user to determine the library rev they are working with.
//
// Short rev history:
//
//  SRHELPER_REV  101  ( 10/15/2003 )  < Changed SrFileExist return, added Rev fn
//  SRHELPER_REV  102  ( 02/25/2006 )  < SrFileLength to take FILE*, remove SrCommit
//  SRHELPER_REV  103  ( 03/20/2006 )  < Added SrFileReadable, SrFileModTime
//          
//------------------------------------------------------------------------------

#define SRHELPER_REV      103

FUNCTYPE( int )  SrHelperRev( int* Rev ) {
        
        *Rev = SRHELPER_REV;
        return( *Rev );
}


//------------------------------------------------------------------------------
// ROUTINE: SrKeyboardNonBlock
// PURPOSE: Ensures non-blocking mode is enabled, allowing SrGetKeyCheck
//          to return even if a key has not been pressed.
//------------------------------------------------------------------------------
FUNCTYPE( int ) SrKeyboardNonBlock( void ) {
        return( OsSetNonBlock( 1 ) );
}

//------------------------------------------------------------------------------
// ROUTINE: SrGetKeyCheck
// PURPOSE: Check for an available keypress character.  If one is there, it is
//          read and returned.  Otherwise, EOF is returned.
//------------------------------------------------------------------------------
FUNCTYPE( int ) SrGetKeyCheck( void ) {
        return( OsGetKey( 0 ) );
}

//------------------------------------------------------------------------------
// ROUTINE: SrGetKeyWait
// PURPOSE: Wait until a keypress character is ready and then return it.
//------------------------------------------------------------------------------
FUNCTYPE( int ) SrGetKeyWait( void ) {
        return( OsGetKey( 1 ) );
}

//------------------------------------------------------------------------------
// ROUTINE: SrFlushPrint
// PURPOSE: Ensure the last printf string is displayed, even if it didn't end
//          with a new line (\n) character.
//------------------------------------------------------------------------------
FUNCTYPE( void ) SrFlushPrint( void ) {
        OsFlushPrint();
}

//------------------------------------------------------------------------------
// ROUTINE: SrSleep
// PURPOSE: Put the calling program to sleep for the specified number of
//          milliseconds.
//------------------------------------------------------------------------------
FUNCTYPE( int ) SrSleep( int ms ) {
        return( OsSleep( ms, 0 ) );
}

//------------------------------------------------------------------------------
// ROUTINE: SrWait
// PURPOSE: Wait for the specified number of milliseconds usually by putting
//          the calling program to sleep.
//------------------------------------------------------------------------------
FUNCTYPE( int ) SrWait( int ms ) {
        return( OsSleep( ms, 1 ) );
}

//------------------------------------------------------------------------------
// ROUTINE: SrFileExist
// PURPOSE: Test to see if the specified file exists.
//          The return value changed as of 10/15/2003.  
//          Now it is 1 when the file exists and 0 otherwise.
//------------------------------------------------------------------------------

#if !defined F_OK
#define F_OK 0x00
#endif

FUNCTYPE( int )  SrFileExist( const char *FileName ) {
        if ( OsAccess( FileName, F_OK ) == 0 )
                return( 1 );
        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: SrFileReadable
// PURPOSE: Test to see if the specified file is readable.
//          It returns 1 when the file can be read and 0 otherwise.
//------------------------------------------------------------------------------

#if !defined R_OK
#define R_OK 0x04
#endif

FUNCTYPE( int )  SrFileReadable( const char *FileName ) {
        if ( OsAccess( FileName, R_OK ) == 0 )
                return( 1 );
        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: SrFileLength
// PURPOSE: Returns the length of the specified file in bytes.
//------------------------------------------------------------------------------
FUNCTYPE( long ) SrFileLength( FILE *fp ) {

        struct stat buffer;

        if ( fstat( OsFileno( fp ), &buffer ) == -1 ) {
                printf( "Could not get file stats\n" );
                return( -1 );
                }

        return( buffer.st_size );
}

//------------------------------------------------------------------------------
// ROUTINE: SrFileModTime
// PURPOSE: Returns the last modification time of the specified file.
//------------------------------------------------------------------------------
FUNCTYPE( long ) SrFileModTime( FILE *fp ) {

        struct stat buffer;

        if ( fstat( OsFileno( fp ), &buffer ) == -1 ) {
                printf( "Could not get file stats\n" );
                return( -1 );
                }

        return( buffer.st_mtime );
}

//------------------------------------------------------------------------------
// ROUTINE: SrStringCopy
// PURPOSE: This function copies up to MaxChar-1 characters from SrcStr to 
//          DestStr, ensuring that the last character in the destination string
//          is a null.  It returns the total number of characters in the
//          destination string, including the terminating null.
//          *** Be sure MaxChar is <= the dimension of DestStr. ***
//------------------------------------------------------------------------------
FUNCTYPE( int ) SrStringCopy( char *DestStr, const char *SrcStr, int MaxChar ) {

        int i;

        // Copy characters from SrcStr to DestStr, stopping after
        // MaxChar-1 characters or a null at the end of the source
        // string is reached.

        i = 1;
        while ( (i < MaxChar) && (*DestStr++ = *SrcStr++) )
                i++;


        // If the copying stopped because of hitting the character
        // limit, fill the last character with null (if the copy stopped
        // because of reaching the end of the source string, the last
        // character is already null).

        if ( i == MaxChar )
                *DestStr = '\0';

        return( i );
}



//------------------------------------------------------------------------------
// OS DEPENDENT HELPER FUNCTIONS:
//
// These functions help hide OS dependencies from the rest of
// the library by providing a common interface.  The functions
// are repeated once for each supported OS, but the SROS_xxxxx
// defines select only one set to be compiled in to the code.
// 
// Supported OS's include
//
//     SROS_WINXP  Windows XP
//     SROS_WIN2K  Windows 2000
//     SROS_WIN9X  Windows 95, 98, and ME
//     SROS_MSDOS  MS Dos
//     SROS_LINUX  Linux 2.6.18-1.2798.fc6 kernel (Fedora Core 6)
//
//
//------------------------------------------------------------------------------


#if defined( SROS_WINXP ) || defined( SROS_WIN2K )

//------------------------------------------------------------------------------
// ROUTINE: OsSetNonBlock (Windows version)
// PURPOSE: Sets non-blocking mode which allows testing for a keypress.
//------------------------------------------------------------------------------
SRLOCAL int OsSetNonBlock( int value ) {

        // Win2K/XP is always in non-blocking mode (value=1)
        //   value = 0 blocks reads until a character is ready
        //   value = 1 allow reads to return immediately
        //   returns -1 for failure, 0 for success

        if (value == 1)
                return( 0 );
        else
                return( -1 );
}

//------------------------------------------------------------------------------
// ROUTINE: OsGetKey (Windows version)
// PURPOSE: Test for a keypress and return it if one is ready.  The wait
//          argument indicates what to do if a keypress is not immediately
//          available.  If wait is true, then wait until one is ready.
//          Otherwise, just return right away.
//------------------------------------------------------------------------------
SRLOCAL int OsGetKey( int wait ) {

        int c;

        // Get a ready keypress or wait until one is ready if requested.

        c = EOF;
        if ( _kbhit() || wait ) {
                c = _getch();
                if (c == 0xD) printf("\n");  // add newline if char was ENTER
                if (c == 0x00 || c == 0xE0 ) // deal with extended key (fn,->)
                        c = _getch();         // c has scan code
                }
        
        return( c );
}

//------------------------------------------------------------------------------
// ROUTINE: OsFlushPrint (Windows version)
// PURPOSE: Ensures that printf statements that don't end with a carriage return
//          (ie \n) are output immediately rather then buffered for later output.
//------------------------------------------------------------------------------
SRLOCAL void OsFlushPrint( void ) {
        
        // Win2K/XP automatically flushes printf.
}

//------------------------------------------------------------------------------
// ROUTINE: OsSleep (Windows version)
// PURPOSE: Sleeps for the specified number of milliseconds.  This usually 
//          allows other processes access to the CPU.  The DOS version just
//          returns immediately unless the hardwait parameter is true.  In which
//          case it busy waits instead.  The return value is 0 for success and
//          -1 for error.
//------------------------------------------------------------------------------
SRLOCAL int OsSleep( int ms, int hardwait ) {

        // Sleep for specifed number of milliseconds.
        

        if ( ms <= 0 )
                return( 0 );
        
        Sleep( ms );
        return( 0 );
}



#elif defined( SROS_WIN9X ) || defined( SROS_MSDOS )

//------------------------------------------------------------------------------
// ROUTINE: OsSetNonBlock (DOS version)
// PURPOSE: Sets non-blocking mode which allows testing for a keypress.
//------------------------------------------------------------------------------
SRLOCAL int OsSetNonBlock( int value ) {

        // Win9x/MsDOS are always in non-blocking mode (value=1)
        //   value = 0 blocks reads until a character is ready
        //   value = 1 allow reads to return immediately
        //   returns -1 for failure, 0 for success

        if (value == 1)
                return( 0 );
        else
                return( -1 );
}

//------------------------------------------------------------------------------
// ROUTINE: OsGetKey (DOS version)
// PURPOSE: Test for a keypress and return it if one is ready.  The wait
//          argument indicates what to do if a keypress is not immediately
//          available.  If wait is true, then wait until one is ready.
//          Otherwise, just return right away.
//------------------------------------------------------------------------------
SRLOCAL int OsGetKey( int wait ) {

        int c;

        // Get a ready keypress or wait until one is ready if requested.

        c = EOF;
        if ( _kbhit() || wait ) {
                c = _getch();
                if (c == 0xD) printf("\n");  // add newline if char was ENTER
                if (c == 0x00 || c == 0xE0 ) // deal with extended key (fn,->)
                        c = _getch();        // c has scan code
                }
        
        return( c );
}

//------------------------------------------------------------------------------
// ROUTINE: OsFlushPrint (DOS version)
// PURPOSE: Ensures that printf statements that don't end with a carriage return
//          (ie \n) are output immediately rather then buffered for later output.
//------------------------------------------------------------------------------
SRLOCAL void OsFlushPrint( void ) {
        
        // Win9x/MsDOS automatically flushes printf.
}

#if defined ( SROS_WIN9X )
//------------------------------------------------------------------------------
// ROUTINE: OsSleep (Windows95/98 version)
// PURPOSE: Sleeps for the specified number of milliseconds.  This usually 
//          allows other processes access to the CPU.  The DOS version just
//          returns immediately unless the hardwait parameter is true.  In which
//          case it busy waits instead.  The return value is 0 for success and
//          -1 for error.
//------------------------------------------------------------------------------
SRLOCAL int OsSleep( int ms, int hardwait ) {

        // Sleep for specifed number of milliseconds.
        // Returns 0 for success, -1 for error.
        
        if ( ms <= 0 )
                return( 0 );

        Sleep( ms );

        return( 0 );
}
#elif defined( SROS_MSDOS )
//------------------------------------------------------------------------------
// ROUTINE: OsSleep (DOS version)
// PURPOSE: Sleeps for the specified number of milliseconds.  This usually 
//          allows other processes access to the CPU.  The DOS version just
//          returns immediately unless the hardwait parameter is true.  In which
//          case it busy waits instead.  The return value is 0 for success and
//          -1 for error.
//------------------------------------------------------------------------------
SRLOCAL int OsSleep( int ms, int hardwait ) {

        struct _timeb timebuffer;
        unsigned short endms;
        time_t endsec;

        // Sleep for specifed number of milliseconds.
        // Returns 0 for success, -1 for error.
        
        if ( ms <= 0 )
                return( 0 );

        // No sleep function is available for MSDOS, so just
        // return.  Unless the user has requested that they
        // really want to wait, even if it means spinning in
        // a loop wasting CPU cycles.
        
        if ( hardwait ) {
                
                _ftime( &timebuffer );
                endsec = timebuffer.time;
                endms  = timebuffer.millitm + ms;
                
                while (endms > 1000) {
                        endms  -= 1000;
                        endsec += 1;
                        }
                
                while (timebuffer.time < endsec)
                        _ftime( &timebuffer );
                
                while ( (timebuffer.time == endsec) &&
                        (timebuffer.millitm < endms) )
                        _ftime( &timebuffer );
                }

                
//                ms *= 1000; // Convert ms to us
//                while ( ms-- )
//                        _asm {  mov  dx,0x379  // Read from status port
//                                in   al,dx     // takes about 1 microsecond
//                                }

        return( 0 );
}
#endif // SROS_WIN9X or SROS_MSDOS for sleep function difference



#elif defined( SROS_LINUX )

//------------------------------------------------------------------------------
// ROUTINE: OsSetNonBlock (Linux version)
// PURPOSE: Sets non-blocking mode which allows testing for a keypress.
//------------------------------------------------------------------------------
SRLOCAL int OsSetNonBlock( int value ) {

        // Set non-blocking mode for keyboard reads
        //   value = 0 blocks reads until a character is ready
        //   value = 1 allow reads to return immediately
        //   returns -1 for failure, 0 for success

        int oldflags, result;

        oldflags = fcntl( STDIN_FILENO, F_GETFL, 0 );

        if (oldflags == -1)
                return( -1 );

        if (value != 0)
                oldflags |= O_NONBLOCK;
        else
                oldflags &= ~O_NONBLOCK;

       result = fcntl( STDIN_FILENO, F_SETFL, oldflags );

       return( result );
}

//------------------------------------------------------------------------------
// ROUTINE: OsGetKey (Linux version)
// PURPOSE: Test for a keypress and return it if one is ready.  The wait
//          argument indicates what to do if a keypress is not immediately
//          available.  If wait is true, then wait until one is ready.
//          Otherwise, just return right away.
//------------------------------------------------------------------------------
SRLOCAL int OsGetKey( int wait ) {

        int c;

        // Get a ready keypress or wait until one is ready if requested.

        // Notes:
        //
        // This version assumes that non-blocking reads have been
        // enabled.  Otherwise, the first getchar will always wait until
        // a character is ready.
        //
        // Getchar requires an ENTER after the character before it will
        // start processing the character.


        c = getchar();
        while ( (c == EOF) && wait )
                c = getchar();

        return( c );
}

//------------------------------------------------------------------------------
// ROUTINE: OsFlushPrint (Linux version)
// PURPOSE: Ensures that printf statements that don't end with a carriage return
//          (ie \n) are output immediately rather then buffered for later output.
//------------------------------------------------------------------------------
SRLOCAL void OsFlushPrint( void ) {
        
        fflush( STDIN_FILENO );
}

//------------------------------------------------------------------------------
// ROUTINE: OsSleep (Linux version)
// PURPOSE: Sleeps for the specified number of milliseconds.  This usually 
//          allows other processes access to the CPU.  The DOS version just
//          returns immediately unless the hardwait parameter is true.  In which
//          case it busy waits instead.  The return value is 0 for success and
//          -1 for error.
//------------------------------------------------------------------------------
SRLOCAL int OsSleep( int ms, int hardwait ) {

        // Sleep for specifed number of milliseconds.
        // Returns 0 for success, -1 for error.

        struct timespec twait, tleft;
        
        if ( ms <= 0 )
                return( 0 );

        
        // Linux nanosleep takes structure with seconds and nanoseconds.
        
        tleft.tv_sec  = tleft.tv_nsec = 0;
        twait.tv_sec  = (long)(ms / 1000);
        twait.tv_nsec = (ms - (twait.tv_sec*1000)) * 1000000;
        
        return( nanosleep( &twait, &tleft ) );
}

#endif // SROS_xxxxx   End of OS dependent helper functions

