/* FILE: meter.c                Copyright (c), Symmetric Research, 2003-2007

Command line program to demonstrate meter style acquisition with text
output.

When running, typing a "t" will act as a trigger request and save the
next data sample for each channel to an output file.

Parameters like sampling rate and channel calibrations are defined
constants below.  Recompile if you change these values.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "parxch.h"
#include "srhelper.h"



/* Defines
 *
 *
 */

#define MAXNAME           20
#define PAGEMAX           22
#define SAMPLEMAX        999
#define DISPLAY_DECCOUNTS  0
#define DISPLAY_HEXCOUNTS  1
#define DISPLAY_VOLTS      2

#define HEADERMSG1 "\n         0\n\n"
#define HEADERMSG4 "\n         0        1        2        3\n\n"
#define HEADERMSG8 "\n         0        1        2        3        4        5        6        7\n\n"




/* Global variables
 *
 *
 */

FILE *fp;
char DriverName[MAXNAME], *XchModelName, *PortModeName, TmpStr[256];
int  DisplayMode, XchModel, PortMode, NeedHelp, AnalogChannels;
int  OutputToFile;





/* PARxCH SAMPLING RATE CONFIGURATION:

These parameters set the sampling rate and averaging.  Setting
SPS = 50.0 Hz and then averaging over NAVERAGE = 50 points gives
an effective sampling rate of 1 Hz.

*/

#define SPS        50.0         // samples per second
#define NAVERAGE   50



        


/* VOLTAGE CALIBRATION INFO:

When displaying the output as counts, the acquired 24 bit data is sign
extended to 32 bits and shown in hexadecimal format.  Positive values
will have 00 as the top two hex digits, while negative values will
have FF.

When displaying the output as volts, a counts to voltage conversion
must be performed.  To perform this conversion, the system must be
calibrated.

The formulas here do a simple straight line conversion.  Given the
counts at two different voltages, the straight line through the two
points is computed.

This of course could be enhanced by computing a least squares fit
line to many sampled points, but the two point formula here works
fairly well and is easy to implement.  Because the system has eight
independent A/D converters, a calibration formula is carried for each
channel.

To calibrate your system, give each channel a known voltage, V1, and
running in counts mode, record the counts for each channel.  Repeat
for a second voltage V2.  Enter the results in the defined constants
below and recompile.  The quality of the result of course depends on
the precision and stability of your voltage references V1, and V2.

*/

#define V1                      (double)-5.0000

#define C1_CHAN0                ((long)0xFFE3E698)
#define C1_CHAN1                ((long)0xFFE3E698)
#define C1_CHAN2                ((long)0xFFE3E698)
#define C1_CHAN3                ((long)0xFFE3E698)
#define C1_CHAN4                ((long)0xFFE3E698)
#define C1_CHAN5                ((long)0xFFE3E698)
#define C1_CHAN6                ((long)0xFFE3E698)
#define C1_CHAN7                ((long)0xFFE3E698)

#define V2                      (double)+5.0000

#define C2_CHAN0                ((long)0x0020EA61)
#define C2_CHAN1                ((long)0x0020EA61)
#define C2_CHAN2                ((long)0x0020EA61)
#define C2_CHAN3                ((long)0x0020EA61)
#define C2_CHAN4                ((long)0x0020EA61)
#define C2_CHAN5                ((long)0x0020EA61)
#define C2_CHAN6                ((long)0x0020EA61)
#define C2_CHAN7                ((long)0x0020EA61)



// slopes for each channel ...

#define SL_CHAN0                (V2-V1)/((double)(C2_CHAN0 - C1_CHAN0))
#define SL_CHAN1                (V2-V1)/((double)(C2_CHAN1 - C1_CHAN1))
#define SL_CHAN2                (V2-V1)/((double)(C2_CHAN2 - C1_CHAN2))
#define SL_CHAN3                (V2-V1)/((double)(C2_CHAN3 - C1_CHAN3))
#define SL_CHAN4                (V2-V1)/((double)(C2_CHAN4 - C1_CHAN4))
#define SL_CHAN5                (V2-V1)/((double)(C2_CHAN5 - C1_CHAN5))
#define SL_CHAN6                (V2-V1)/((double)(C2_CHAN6 - C1_CHAN6))
#define SL_CHAN7                (V2-V1)/((double)(C2_CHAN7 - C1_CHAN7))


// formulas for each channel ...

#define CtoV_CHAN0( C)          ( SL_CHAN0*(C - C1_CHAN0) + V1 )
#define CtoV_CHAN1( C)          ( SL_CHAN1*(C - C1_CHAN1) + V1 )
#define CtoV_CHAN2( C)          ( SL_CHAN2*(C - C1_CHAN2) + V1 )
#define CtoV_CHAN3( C)          ( SL_CHAN3*(C - C1_CHAN3) + V1 )
#define CtoV_CHAN4( C)          ( SL_CHAN4*(C - C1_CHAN4) + V1 )
#define CtoV_CHAN5( C)          ( SL_CHAN5*(C - C1_CHAN5) + V1 )
#define CtoV_CHAN6( C)          ( SL_CHAN6*(C - C1_CHAN6) + V1 )
#define CtoV_CHAN7( C)          ( SL_CHAN7*(C - C1_CHAN7) + V1 )




/* Function Prototypes
 *
 *
 */

void  Display( int DisplayMode, long counts[], int page, int sample );
void  ProcessCommandLine( int argc, char **argv );
char *StrDisplayMode( int PassedDisplayMode );
char *StrXchModel( int PassedXchModel );
char *StrPortMode( int PassedPortMode );
int   GetAnalogChannels( int PassedXchModel );
void  KeyCheck( DEVHANDLE handle );
void  Usage( char *msg );
void  PrintMsg( char *MsgArray[] );



/* MESSAGES:
 *
 *
 *
 */

char *UsageMsg[] = {
  "USAGE: meter DisplayMode [DriverName] [XchModel] [PortMode] [/?]            ",
  "                                                                            ",
  "  DisplayMode = volts, counts, hexcounts                                    ",
  "  DriverName  = SrParXch0, SrParXch1, SrParXch2        default SrParXch0    ",
  "  XchModel    = PAR1CH, PAR4CH, PAR8CH                 default from setmodel",
  "  PortMode    = bpp, epp, ecp/bpp, ecp/epp             default ecp/bpp      ",
  "  /?          = show this usage message                                     ",
  "                                                                            ",
  "EXAMPLES: meter volts                                                       ",
  "          meter counts SrParXch0 PAR4CH ecp/bpp                             ",
  NULL
  };


char *CalibrationMsg[] = {
  "                                                                         ",
  "NOTE: Meter has been compiled with default calibration values for        ",
  "converting counts to volts, see meter.c for details.                     ",
  "                                                                         ",
  NULL
  };

char *SuccessCountsMsg[] = {
  "SUCCESS ...                                                              ",
  "                                                                         ",
  "Acquiring and displaying analog values as COUNTS, press ENTER to quit ...",
  NULL
  };

char *SuccessHexCountsMsg[] = {
  "SUCCESS ...                                                              ",
  "                                                                         ",
  "Acquiring and displaying analog values as HEX COUNTS, press ENTER to quit ...",
  NULL
  };

char *SuccessVoltsMsg[] = {
  "SUCCESS ...                                                              ",
  "                                                                         ",
  "Acquiring and displaying analog values as VOLTS, press ENTER to quit ...",
  NULL
  };





/* MAIN:
 *
 * Initialize the PARxCH and start acquisition.
 *
 *
 *
 */

int main( int argc, char *argv[] ) {

        int    i, ipage, isample, naverage, npts, ErrorCode;
        long   counts[PARXCH_MAX_ANALOG_CHANNELS];
        long   averages[PARXCH_MAX_ANALOG_CHANNELS];
        double ActualSps;
        DEVHANDLE ParXchHandle;


        // Announce program version.

        printf( "PARxCH meter [/?]: (c) SR 2007 : Rev {01/08/07}\n" );


        // Get command line arguments

        ProcessCommandLine( argc, argv );



        // Open output file

        fp = fopen( "METER.OUT", "w" );
        if (!fp) {
                printf("Could not open output file METER.OUT\n" );
                exit( 0 );
                }

        // The display function will also write the current values to
        // the output file when this variable is set to 1.
        
        OutputToFile = 0;  



        // Set keyboard behavior to allow for checking a key.

        if ( SrKeyboardNonBlock() == -1 )
                { printf( "Can't set non-blocking read\n"); exit( 1 ); }


        // Show informative message and wait for a keypress to continue.

        PrintMsg( CalibrationMsg );
        printf( "Press ENTER to continue ..." );  SrFlushPrint();
        SrGetKeyWait( );


        
        

        // Initialize the PARxCH.

        printf( "\nOpening PARxCH A/D board with:\n" );
        printf( "   DriverName = %s\n",   DriverName );
        printf( "   XchModel   = %s\n",   XchModelName );
        printf( "   PortMode   = %s\n\n", PortModeName );

        ParXchHandle = ParXchOpen(
                                  DriverName,
                                  XchModel,
                                  PortMode,
                                  SPS,

                                  &ActualSps,
                                  &ErrorCode
                                 );

        if ( ParXchHandle == BAD_DEVHANDLE ) { 
                printf( "ERROR: Could not open PARxCH. (Err=%s)\n\n",
                          PARXCH_ERROR_MSG[ErrorCode] );
                exit( 1 );
                }

        
        // Announce success

        if ( DisplayMode == DISPLAY_DECCOUNTS )
                PrintMsg( SuccessCountsMsg );
        else if ( DisplayMode == DISPLAY_HEXCOUNTS )
                PrintMsg( SuccessHexCountsMsg );
        else
                PrintMsg( SuccessVoltsMsg );




        // Initialize the naverage count, and averages for each channel.

        ipage    = 0;
        isample  = 0;
        naverage = 0;

        for ( i = 0 ; i < AnalogChannels ; i++ )

                averages[i] = 0;
        


        
        // Start acquisition

        ParXchStart( ParXchHandle );

polling_loop:



        // If data is ready ...
        
        if ( ParXchReady( ParXchHandle ) ) {


                // Read data.

                npts = ParXchReadData(
                                       ParXchHandle,
                                       counts,
                                       AnalogChannels,
                                       &ErrorCode
                                     );



                // Check for an error.

                if ( (npts != AnalogChannels)       ||
                     (ErrorCode != PARXCH_ERROR_NONE) ){
                        printf( "\nReadData %d points, error %s\n",
                                 npts,
                                 PARXCH_ERROR_MSG[ErrorCode]
                                );
                        printf( "NOTE: Spcifiying the wrong PARxCH model may be reported as a FIFO SANITY error.\n" ); 
                        ParXchClose( ParXchHandle );
                        exit( 1 );
                        }

                

                // Add in the new sample from each channel.

                for ( i = 0 ; i < AnalogChannels ; i++ )

                        averages[i] += counts[i];



                // Increment the number of samples averaged.

                naverage++;



                // If the last sample has been averaged, normalize and display.

                if ( naverage >= NAVERAGE ) {


                        // Normalize.

                        for ( i = 0 ; i < AnalogChannels ; i++ )

                                averages[i] /= NAVERAGE;


                        // Display the count averages for each channel.

                        Display( DisplayMode, averages, ipage, isample );



                        // Reset the naverage count and averages.

                        naverage = 0;

                        for ( i = 0 ; i < AnalogChannels ; i++ )

                                averages[i] = 0;


                        // Update sample count and page count, wrap if needed

                        isample++;
                        if (isample > SAMPLEMAX)
                                isample++;
                        ipage++;
                        if (ipage > PAGEMAX)
                                ipage = 0;

                        
                        }  // end if naverage >= NAVERAGE

                
                }  // end if ParXchReady


        
        // Check for a keypress to quit.

        KeyCheck( ParXchHandle );
        


        // Wait a little, then continue looping.

        SrSleep( 15 ); // ms

        goto polling_loop;
        
        return( 0 );
}




/* DISPLAY FUNCTION:
 *
 *
 *
 *
 *
 */

void Display( int DisplayMode, long counts[], int page, int sample ) {

        int i, pt;

        // Print a header line at the top of the "page"

        if (page == 0) {
                if (XchModel == PARXCH_MODEL_PAR1CH)
                        printf( HEADERMSG1 );
                else if (XchModel == PARXCH_MODEL_PAR4CH)
                        printf( HEADERMSG4 );
                else // if (XchModel == PARXCH_MODEL_PAR8CH)
                        printf( HEADERMSG8 );
                }



        // Display the data from all analog channels.

        pt = sample%1000; // Make index value that will fit in 3 positions

        if ( DisplayMode == DISPLAY_DECCOUNTS ) {

                printf( "(%03d)", pt );
                for ( i = 0 ; i < AnalogChannels ; i++ )
                        printf( " %8ld", counts[i] );
                printf( "\n" );

                if ( OutputToFile == 1 ) {
                        fprintf( fp, "(%03d)", pt );
                        for ( i = 0 ; i < AnalogChannels ; i++ )
                                fprintf( fp, " %8ld", counts[i] );
                        fprintf( fp, "\n" );
                        OutputToFile = 0;
                        }
                }

        else if ( DisplayMode == DISPLAY_HEXCOUNTS ) {

                printf( "(%03d)", pt );
                for ( i = 0 ; i < AnalogChannels ; i++ )
                        printf( " %08lX", counts[i] );
                printf( "\n" );

                if ( OutputToFile == 1 ) {
                        fprintf( fp, "(%03d)", pt );
                        for ( i = 0 ; i < AnalogChannels ; i++ )
                                fprintf( fp, " %08lX", counts[i] );
                        fprintf( fp, "\n" );
                        OutputToFile = 0;
                        }
                }

        else { // DisplayMode == DISPLAY_VOLTS

                if (XchModel == PARXCH_MODEL_PAR1CH) {
                        printf( "(%03d) %+08.4f\n",
                                  pt,
                                  CtoV_CHAN0( counts[0] )
                              );

                        if ( OutputToFile == 1 ) {
                                fprintf( fp, "(%03d) %+08.4f\n",
                                          pt,
                                          CtoV_CHAN0( counts[0] )
                                      );
                                OutputToFile = 0;
                                }
                        }
                                
                else if (XchModel == PARXCH_MODEL_PAR4CH) {
                        printf( "(%03d) %+08.4f %+08.4f %+08.4f %+08.4f\n",
                                  pt,
                                  CtoV_CHAN0( counts[0] ),
                                  CtoV_CHAN1( counts[1] ),
                                  CtoV_CHAN2( counts[2] ),
                                  CtoV_CHAN3( counts[3] )
                              );
                        if ( OutputToFile == 1 ) {
                                fprintf( fp, "(%03d) %+08.4f %+08.4f %+08.4f %+08.4f\n",
                                         pt,
                                         CtoV_CHAN0( counts[0] ),
                                         CtoV_CHAN1( counts[1] ),
                                         CtoV_CHAN2( counts[2] ),
                                         CtoV_CHAN3( counts[3] )
                                       );
                                OutputToFile = 0;
                                }
                        }

                else if ( XchModel == PARXCH_MODEL_PAR8CH) {
                        printf( "(%03d) %+08.4f %+08.4f %+08.4f %+08.4f %+08.4f %+08.4f %+08.4f %+08.4f\n",
                                  pt,
                                  CtoV_CHAN0( counts[0] ),
                                  CtoV_CHAN1( counts[1] ),
                                  CtoV_CHAN2( counts[2] ),
                                  CtoV_CHAN3( counts[3] ),
                                  CtoV_CHAN4( counts[4] ),
                                  CtoV_CHAN5( counts[5] ),
                                  CtoV_CHAN6( counts[6] ),
                                  CtoV_CHAN7( counts[7] )
                              );
                        if ( OutputToFile == 1 ) {
                                fprintf( fp, "(%03d) %+08.4f %+08.4f %+08.4f %+08.4f %+08.4f %+08.4f %+08.4f %+08.4f\n",
                                         pt,
                                         CtoV_CHAN0( counts[0] ),
                                         CtoV_CHAN1( counts[1] ),
                                         CtoV_CHAN2( counts[2] ),
                                         CtoV_CHAN3( counts[3] ),
                                         CtoV_CHAN4( counts[4] ),
                                         CtoV_CHAN5( counts[5] ),
                                         CtoV_CHAN6( counts[6] ),
                                         CtoV_CHAN7( counts[7] )
                                       );
                                OutputToFile = 0;
                                }
                        }
                else
                        printf( "Unknown XchModel = %d\n", XchModel );

                }

}




/* COMMAND LINE PROCESSING FUNCTION:
 *
 *
 */


#define CMD_DONE    -1
#define CMD_HELP     0
#define CMD_DRIVER   1
#define CMD_MODEL    2
#define CMD_PORT     3
#define CMD_DISPLAY  4
#define CMD_MAX      5

struct CommandLineOption {
        int   Category;
        char *Name;
        char *ParmChar;
        int  *ParmInt;
        int   Value;
        } *Opt, OptionsTable[] = {

  CMD_HELP,   "/?",           NULL,       &NeedHelp,    1,

  CMD_DRIVER,  "SrParXch0",   DriverName, NULL,         0,
  CMD_DRIVER,  "SrParXch1",   DriverName, NULL,         0,
  CMD_DRIVER,  "SrParXch2",   DriverName, NULL,         0,

  CMD_MODEL,   "PAR1CH",      NULL,       &XchModel,    PARXCH_MODEL_PAR1CH,
  CMD_MODEL,   "PAR4CH",      NULL,       &XchModel,    PARXCH_MODEL_PAR4CH,
  CMD_MODEL,   "PAR8CH",      NULL,       &XchModel,    PARXCH_MODEL_PAR8CH,

  CMD_PORT,    "bpp",         NULL,       &PortMode,    PARXCH_PORT_MODE_BPP,
  CMD_PORT,    "epp",         NULL,       &PortMode,    PARXCH_PORT_MODE_EPP,
  CMD_PORT,    "ecp/bpp",     NULL,       &PortMode,    PARXCH_PORT_MODE_ECP_BPP,
  CMD_PORT,    "ecp/epp",     NULL,       &PortMode,    PARXCH_PORT_MODE_ECP_EPP,

  CMD_DISPLAY, "counts",      NULL,       &DisplayMode, DISPLAY_DECCOUNTS,
  CMD_DISPLAY, "hexcounts",   NULL,       &DisplayMode, DISPLAY_HEXCOUNTS,
  CMD_DISPLAY, "volts",       NULL,       &DisplayMode, DISPLAY_VOLTS,
  
  CMD_DONE,     NULL,         NULL,       NULL,         0,

        };


void ProcessCommandLine( int argc, char **argv ) {

        int i, CmdFilled[CMD_MAX];


        // Set defaults

        strncpy( DriverName, PARXCH_0, MAXNAME );
        DisplayMode = DISPLAY_VOLTS;
        XchModel    = PARXCH_MODEL_UNKNOWN;
        PortMode    = PARXCH_PORT_MODE_ECP_BPP;
        NeedHelp    = 0;

        for ( i = 0 ; i < CMD_MAX ; i++ )
                CmdFilled[i] = 0;
        

        // Verify number of arguments (one to 4)

        if ( argc < 2 || argc > 5 )
                Usage( "wrong number of arguments" );


        // Process arguments

        while ( --argc ) {

                argv++; // Move to next argument (1st time skips program name)
                
                // Check all names in the OptionsTable for a match.

                Opt = OptionsTable;
                while ( Opt->Name && strcmp( *argv, Opt->Name ) )
                                Opt++;


                // Argument did not match a valid option.

                if ( !Opt->Name ) {
                        sprintf( TmpStr, "unknown option %s", *argv );
                        Usage( TmpStr );
                        }


                // Check for duplicate argument.

                if ( CmdFilled[ Opt->Category ] == 1 )
                        Usage( "more than one option per category" );
                else
                        CmdFilled[ Opt->Category ] = 1;

                
                // Fill in known option.

                if ( Opt->ParmInt )
                        *Opt->ParmInt = Opt->Value;
                
                else if ( Opt->ParmChar )
                        SrStringCopy( Opt->ParmChar, Opt->Name, MAXNAME );

                else
                        Usage( "option of unknown type" );

                } // end while argc



        // Check for help request

        if ( NeedHelp == 1 )
                Usage( "?" );



        // Finalize selection of specific PARxCH model

        if (XchModel == PARXCH_MODEL_UNKNOWN) // Not given by user on command line
                XchModel = ParXchGetDefaultModel( DriverName );

        AnalogChannels = GetAnalogChannels( XchModel );
        XchModelName   = StrXchModel( XchModel );
        PortModeName   = StrPortMode( PortMode );

        
        return;
}








/* HELPER FUNCTIONS:
 *
 *
 *
 */
        
char *StrDisplayMode( int PassedDisplayMode ) {

        if (PassedDisplayMode == DISPLAY_DECCOUNTS)
                return( "COUNTS" );
        else if (PassedDisplayMode == DISPLAY_HEXCOUNTS)
                return( "HEX COUNTS" );
        else if (PassedDisplayMode == DISPLAY_VOLTS)
                return( "VOLTS" );
        else 
                return( "unknown" );
}

char *StrXchModel( int PassedXchModel ) {

        if (PassedXchModel == PARXCH_MODEL_PAR1CH)
                return( PAR1CH_MODELNAME );
        else if (PassedXchModel == PARXCH_MODEL_PAR4CH)
                return( PAR4CH_MODELNAME );
        else if (PassedXchModel == PARXCH_MODEL_PAR8CH)
                return( PAR8CH_MODELNAME );
        else 
                return( "unknown" );
}

char *StrPortMode( int PassedPortMode ) {

        if (PassedPortMode == PARXCH_PORT_MODE_BPP)
                return( "BPP" );
        else if (PassedPortMode == PARXCH_PORT_MODE_EPP)
                return( "EPP" );
        else if (PassedPortMode == PARXCH_PORT_MODE_ECP_BPP)
                return( "ECP/BPP" );
        else if (PassedPortMode == PARXCH_PORT_MODE_ECP_EPP)
                return( "ECP/EPP" );
        else 
                return( "unknown" );
}

int GetAnalogChannels( int PassedXchModel ) {

        if (PassedXchModel == PARXCH_MODEL_PAR1CH)
                return( PAR1CH_ANALOG_CHANNELS );
        else if (PassedXchModel == PARXCH_MODEL_PAR4CH)
                return( PAR4CH_ANALOG_CHANNELS );
        else if (PassedXchModel == PARXCH_MODEL_PAR8CH)
                return( PAR8CH_ANALOG_CHANNELS );
        else 
                return( PARXCH_MAX_ANALOG_CHANNELS );
}

void KeyCheck( DEVHANDLE ParXchHandle ) {

        int c;

        // Check for a keypress, quitting if found.
        
        c = SrGetKeyCheck( );

        if ( c == 't' ) {
                OutputToFile = 1;
                return;
                }

        if ( c != EOF ) {

                ParXchStop( ParXchHandle );
                ParXchClose( ParXchHandle );
                fclose( fp );
                exit( 0 );
                }
}


void Usage( char *msg ) {

        // Print a command line usage message and quit.

        // Select error line depending on 
        // 1) no message, 2) help requested, or 3) message given

        if ( !msg )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( msg, "?" ) == 0 )
                printf( "\n" );

        else
                printf( "\nERROR: Command line usage (%s).\n\n", msg );

        PrintMsg( UsageMsg );
        exit( 1 );
}

void PrintMsg( char *MsgArray[] ) {

        int i;
        
        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );
}
